from orangecanvas.localization import Translator  # pylint: disable=wrong-import-order
_tr = Translator("Orange", "biolab.si", "Orange")
del Translator
import concurrent.futures
from dataclasses import dataclass
from typing import (
    Optional, Union, Sequence, List, TypedDict, Tuple, Any, Container
)

import numpy as np
from scipy.sparse import issparse

from AnyQt.QtWidgets import (
    QTableView, QHeaderView, QApplication, QStyle, QStyleOptionHeader,
    QStyleOptionViewItem
)
from AnyQt.QtGui import QColor, QClipboard, QPainter
from AnyQt.QtCore import (
    QTimer, Qt, QSize, QMetaObject, QItemSelectionModel, QModelIndex, QRect,
    QAbstractProxyModel, QObject, QEvent
)
from AnyQt.QtCore import Slot

from orangewidget.gui import OrangeUserRole

import Orange.data
from Orange.data.table import Table
from Orange.data.sql.table import SqlTable

from Orange.widgets import gui
from Orange.widgets.data.utils.models import RichTableModel, TableSliceProxy
from Orange.widgets.utils.itemdelegates import TableDataDelegate
from Orange.widgets.utils.tableview import table_selection_to_mime_data
from Orange.widgets.utils.widgetpreview import WidgetPreview
from Orange.widgets.widget import OWWidget, Input, Output, Msg
from Orange.widgets.utils.annotated_data import (create_annotated_table,
                                                 ANNOTATED_DATA_SIGNAL_NAME)
from Orange.widgets.utils.itemmodels import TableModel
from Orange.widgets.utils.state_summary import format_summary_details
from Orange.widgets.utils import disconnected
from Orange.widgets.utils.headerview import HeaderView
from Orange.widgets.data.utils.tableview import RichTableView
from Orange.widgets.data.utils import tablesummary as tsummary
from Orange.widgets.settings import Setting
import os
if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
    from Orange.widgets.orangecontrib.AAIT.utils import widget_positioning, help_management
else:
    from orangecontrib.AAIT.utils import widget_positioning, help_management


SubsetRole = next(OrangeUserRole)


# ---------------------------------------------------------------------------
# Proxy de réordonnancement des lignes
# ---------------------------------------------------------------------------
class RowReorderProxyModel(QAbstractProxyModel):
    """
    Proxy intercalé entre _TableModel et la vue.
    _row_order[i] = index source affiché à la position i.
    Le tri est délégué au source (RichTableModel gère son propre tri interne).
    _row_order est remis à [0..n-1] après chaque tri source.
    """

    def __init__(self, parent=None):
        super().__init__(parent)
        self._row_order: List[int] = []

    def setSourceModel(self, source_model):
        super().setSourceModel(source_model)
        if source_model is not None:
            self._row_order = list(range(source_model.rowCount()))
        else:
            self._row_order = []
        self.beginResetModel()
        self.endResetModel()

    def setRowOrder(self, order: List[int]):
        if self.sourceModel() is None:
            return
        n = self.sourceModel().rowCount()
        if sorted(order) == list(range(n)):
            self.beginResetModel()
            self._row_order = list(order)
            self.endResetModel()

    def getRowOrder(self) -> List[int]:
        return list(self._row_order)

    def moveRow(self, from_pos: int, to_pos: int):
        if from_pos == to_pos:
            return
        self.beginResetModel()
        item = self._row_order.pop(from_pos)
        self._row_order.insert(to_pos, item)
        self.endResetModel()

    def sort(self, column: int, order: Qt.SortOrder = Qt.AscendingOrder):
        src = self.sourceModel()
        if src is None:
            return
        # On délègue au source, puis on remet _row_order à plat.
        # La restauration de l'ordre original est gérée dans le widget
        # en remplaçant le source par un nouveau _TableModel vierge.
        self.beginResetModel()
        src.sort(column, order)
        self._row_order = list(range(src.rowCount()))
        self.endResetModel()

    def rowCount(self, parent=QModelIndex()):
        if parent.isValid() or self.sourceModel() is None:
            return 0
        return len(self._row_order)

    def columnCount(self, parent=QModelIndex()):
        if parent.isValid() or self.sourceModel() is None:
            return 0
        return self.sourceModel().columnCount()

    def mapToSource(self, proxy_index: QModelIndex) -> QModelIndex:
        if not proxy_index.isValid() or self.sourceModel() is None:
            return QModelIndex()
        src_row = self._row_order[proxy_index.row()]
        return self.sourceModel().index(src_row, proxy_index.column())

    def mapFromSource(self, source_index: QModelIndex) -> QModelIndex:
        if not source_index.isValid():
            return QModelIndex()
        try:
            proxy_row = self._row_order.index(source_index.row())
        except ValueError:
            return QModelIndex()
        return self.index(proxy_row, source_index.column())

    def index(self, row, column, parent=QModelIndex()):
        if parent.isValid():
            return QModelIndex()
        return self.createIndex(row, column)

    def parent(self, index=QModelIndex()):
        return QModelIndex()

    def data(self, index: QModelIndex, role=Qt.DisplayRole):
        return self.sourceModel().data(self.mapToSource(index), role)

    def headerData(self, section, orientation, role=Qt.DisplayRole):
        if orientation == Qt.Vertical and self.sourceModel() is not None:
            src_row = self._row_order[section] if section < len(self._row_order) else section
            return self.sourceModel().headerData(src_row, orientation, role)
        if self.sourceModel() is not None:
            return self.sourceModel().headerData(section, orientation, role)
        return None

    def flags(self, index: QModelIndex):
        return self.sourceModel().flags(self.mapToSource(index))


# ---------------------------------------------------------------------------
# Event filter installé sur le VIEWPORT du header vertical
# ---------------------------------------------------------------------------
class RowDragEventFilter(QObject):
    def __init__(self, header: QHeaderView, on_move_callback):
        super().__init__(header)
        self._header = header
        self._on_move = on_move_callback
        self._drag_row: Optional[int] = None
        self._drag_start_y: int = 0
        self._active: bool = False
        self._indicator_row: Optional[int] = None

    def _row_at(self, y: int) -> int:
        return self._header.logicalIndexAt(y)

    def eventFilter(self, obj, event) -> bool:
        t = event.type()

        if t == QEvent.MouseButtonPress and event.button() == Qt.LeftButton:
            row = self._row_at(event.pos().y())
            if row >= 0:
                self._drag_row = row
                self._drag_start_y = event.pos().y()
                self._active = False
                self._indicator_row = None
            return False

        elif t == QEvent.MouseMove and (event.buttons() & Qt.LeftButton):
            if self._drag_row is not None:
                dy = abs(event.pos().y() - self._drag_start_y)
                if dy > 6:
                    self._active = True
                if self._active:
                    target = self._row_at(event.pos().y())
                    if target < 0:
                        target = self._header.count() - 1
                    if target != self._indicator_row:
                        self._indicator_row = target
                        self._header.viewport().update()
                    return True

        elif t == QEvent.MouseButtonRelease and event.button() == Qt.LeftButton:
            if self._drag_row is not None and self._active:
                from_row = self._drag_row
                to_row = self._indicator_row if self._indicator_row is not None else from_row
                self._drag_row = None
                self._active = False
                self._indicator_row = None
                self._header.viewport().update()
                if from_row != to_row:
                    self._on_move(from_row, to_row)
                return True
            self._drag_row = None
            self._active = False
            self._indicator_row = None

        return False


# ---------------------------------------------------------------------------
# Header vertical avec indicateur visuel
# ---------------------------------------------------------------------------
class DraggableVerticalHeader(QHeaderView):
    def __init__(self, parent=None):
        super().__init__(Qt.Vertical, parent)
        self._filter: Optional[RowDragEventFilter] = None
        self.setSectionsClickable(True)
        self.setHighlightSections(True)

    def installDragFilter(self, on_move_callback):
        if self._filter is not None:
            self.viewport().removeEventFilter(self._filter)
        self._filter = RowDragEventFilter(self, on_move_callback)
        self.viewport().installEventFilter(self._filter)

    def paintSection(self, painter: QPainter, rect: QRect, logical_index: int):
        super().paintSection(painter, rect, logical_index)
        if self._filter is None or not self._filter._active:
            return
        if logical_index == self._filter._drag_row:
            painter.save()
            painter.fillRect(rect, QColor(60, 120, 255, 45))
            painter.restore()
        if logical_index == self._filter._indicator_row:
            painter.save()
            pen = painter.pen()
            pen.setColor(QColor(60, 120, 255))
            pen.setWidth(2)
            painter.setPen(pen)
            painter.drawLine(rect.left(), rect.top(), rect.right(), rect.top())
            painter.restore()


class HeaderViewWithSubsetIndicator(HeaderView):
    _IndicatorChar = "\N{BULLET}"

    def paintSection(self, painter: QPainter, rect: QRect, logicalIndex: int) -> None:
        opt = QStyleOptionHeader()
        self.initStyleOption(opt)
        self.initStyleOptionForIndex(opt, logicalIndex)
        model = self.model()
        if model is None:
            return
        opt.rect = rect
        issubset = model.headerData(logicalIndex, Qt.Vertical, SubsetRole)
        style = self.style()
        style.drawControl(QStyle.CE_HeaderSection, opt, painter, self)
        indicator_rect = QRect(rect)
        text_rect = QRect(rect)
        indicator_width = opt.fontMetrics.horizontalAdvance(self._IndicatorChar + " ")
        indicator_rect.setWidth(indicator_width)
        text_rect.setLeft(indicator_width)
        if issubset:
            optindicator = QStyleOptionHeader(opt)
            optindicator.rect = indicator_rect
            optindicator.textAlignment = Qt.AlignCenter
            optindicator.text = self._IndicatorChar
            style.drawControl(QStyle.CE_HeaderLabel, optindicator, painter, self)
        opt.rect = text_rect
        style.drawControl(QStyle.CE_HeaderLabel, opt, painter, self)

    def sectionSizeFromContents(self, logicalIndex: int) -> QSize:
        opt = QStyleOptionHeader()
        self.initStyleOption(opt)
        super().initStyleOptionForIndex(opt, logicalIndex)
        opt.text = self._IndicatorChar + " " + opt.text
        return self.style().sizeFromContents(QStyle.CT_HeaderSection, opt, QSize(), self)


class DataTableView(gui.HScrollStepMixin, RichTableView):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._install_draggable_header()

    def _install_draggable_header(self):
        vheader = DraggableVerticalHeader(self)
        vheader.setSectionsClickable(True)
        self.setVerticalHeader(vheader)

    def setModel(self, model):
        super().setModel(model)
        if not isinstance(self.verticalHeader(), DraggableVerticalHeader):
            self._install_draggable_header()


class _TableDataDelegate(TableDataDelegate):
    DefaultRoles = TableDataDelegate.DefaultRoles + (SubsetRole,)


class SubsetTableDataDelegate(_TableDataDelegate):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.subset_opacity = 0.5

    def paint(self, painter: QPainter, option: QStyleOptionViewItem, index: QModelIndex) -> None:
        issubset = self.cachedData(index, SubsetRole)
        opacity = painter.opacity()
        if not issubset:
            painter.setOpacity(self.subset_opacity)
        super().paint(painter, option, index)
        if not issubset:
            painter.setOpacity(opacity)


class TableBarItemDelegate(SubsetTableDataDelegate, gui.TableBarItem, _TableDataDelegate):
    pass


class _TableModel(RichTableModel):
    SubsetRole = SubsetRole

    def __init__(self, *args, subsets=None, **kwargs):
        super().__init__(*args, **kwargs)
        self._subset = subsets or set()

    def setSubsetRowIds(self, subsetids: Container[int]):
        self._subset = subsetids
        if self.rowCount():
            self.headerDataChanged.emit(Qt.Vertical, 0, self.rowCount() - 1)
            self.dataChanged.emit(
                self.index(0, 0),
                self.index(self.rowCount() - 1, self.columnCount() - 1),
                [SubsetRole],
            )

    def _is_subset(self, row):
        row = self.mapToSourceRows(row)
        try:
            id_ = self.source.ids[row]
        except (IndexError, AttributeError):
            return False
        return int(id_) in self._subset

    def data(self, index: QModelIndex, role=Qt.DisplayRole) -> Any:
        if role == _TableModel.SubsetRole:
            return self._is_subset(index.row())
        return super().data(index, role)

    def headerData(self, section, orientation, role):
        if orientation == Qt.Vertical and role == _TableModel.SubsetRole:
            return self._is_subset(section)
        return super().headerData(section, orientation, role)


@dataclass
class InputData:
    table: Table
    summary: Union[tsummary.Summary, tsummary.ApproxSummary]
    model: TableModel


class _Selection(TypedDict):
    rows: Tuple[int]
    columns: Tuple[int]


_Sorting = List[Tuple[str, int]]


class OWTable(OWWidget):
    name = "Autoshow Data Table"
    description = "View the dataset in a spreadsheet."
    icon = "icons/Table.svg"
    if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
        icon = "icons_dev/Table.svg"
    category = "AAIT - TOOLBOX"
    priority = 50
    keywords = "data table, view"

    class Inputs:
        data = Input("Data", Table, default=True)
        data_subset = Input("Data Subset", Table)
        input_autoshow = Input("AutoShowConfiguration", str, auto_summary=False)

    class Outputs:
        selected_data = Output("Selected Data", Table, default=True)
        annotated_data = Output(ANNOTATED_DATA_SIGNAL_NAME, Table)

    class Warning(OWWidget.Warning):
        missing_sort_columns = Msg(
            ("Cannot restore sorting.\n" + "Missing columns in input table: {}")
        )
        non_sortable_input = Msg(
            ("Cannot restore sorting.\n" + "Input table cannot be sorted due to implementation constraints.")
        )

    str_WidgetPositionning: str = Setting("None")
    buttons_area_orientation = Qt.Vertical

    show_distributions = Setting(False)
    show_attribute_labels = Setting(True)
    select_rows = Setting(True)
    auto_commit = Setting(True)

    color_by_class = Setting(True)
    stored_selection: _Selection = Setting(
        {"rows": [], "columns": []}, schema_only=True
    )
    stored_sort: _Sorting = Setting([], schema_only=True)
    stored_row_order: List[int] = Setting([], schema_only=True)

    settings_version = 1

    def __init__(self):
        super().__init__()
        self.input: Optional[InputData] = None
        self._subset_ids: Optional[set] = None
        self.__pending_selection: Optional[_Selection] = self.stored_selection
        self.__pending_sort: Optional[_Sorting] = self.stored_sort
        self.__have_new_data = False
        self.__have_new_subset = False
        self.dist_color = QColor(220, 220, 220, 255)

        self._reorder_proxy = RowReorderProxyModel()

        info_box = gui.vBox(self.controlArea, "Info")
        self.info_text = gui.widgetLabel(info_box)

        box = gui.vBox(self.controlArea, "Variables")
        self.c_show_attribute_labels = gui.checkBox(
            box, self, "show_attribute_labels",
            "Show variable labels (if present)",
            callback=self._update_variable_labels)

        gui.checkBox(box, self, "show_distributions",
                     'Visualize numeric values',
                     callback=self._on_distribution_color_changed)
        gui.checkBox(box, self, "color_by_class", 'Color by instance classes',
                     callback=self._on_distribution_color_changed)

        box = gui.vBox(self.controlArea, "Selection")
        gui.checkBox(box, self, "select_rows", "Select full rows",
                     callback=self._on_select_rows_changed)

        gui.rubber(self.controlArea)

        gui.button(self.buttonsArea, self, "Restore Original Order",
                   callback=self.restore_order,
                   tooltip="Show rows in the original order",
                   autoDefault=False,
                   attribute=Qt.WA_LayoutUsesWidgetRect)

        gui.button(self.buttonsArea, self, "Reset Row Order",
                   callback=self._reset_row_order,
                   tooltip="Remet les lignes dans l'ordre d'origine",
                   autoDefault=False,
                   attribute=Qt.WA_LayoutUsesWidgetRect)

        gui.auto_send(self.buttonsArea, self, "auto_commit")

        view = DataTableView(sortingEnabled=True)
        view.setItemDelegate(SubsetTableDataDelegate(view))
        view.selectionFinished.connect(self.update_selection)

        if self.select_rows:
            view.setSelectionBehavior(QTableView.SelectRows)

        header = view.horizontalHeader()
        header.setSectionsMovable(True)
        header.setSectionsClickable(True)
        header.setSortIndicatorShown(True)
        header.setSortIndicator(-1, Qt.AscendingOrder)
        header.sortIndicatorChanged.connect(
            self._on_sort_indicator_changed, Qt.UniqueConnection
        )

        self.view = view
        self.mainArea.layout().addWidget(self.view)
        self._update_input_summary()
        widget_positioning.show_and_adjust_at_opening(self, str(self.str_WidgetPositionning))
        QTimer.singleShot(0, lambda: help_management.override_help_action(self))

    def copy_to_clipboard(self):
        self.copy()

    def sizeHint(self):
        return QSize(800, 500)

    @Inputs.data
    def set_dataset(self, data: Optional[Table]):
        if data is not None:
            summary = tsummary.table_summary(data)
            self.input = InputData(
                table=data,
                summary=summary,
                model=_TableModel(data)
            )
            if isinstance(summary.len, concurrent.futures.Future):
                def update(_):
                    QMetaObject.invokeMethod(self, "_update_info", Qt.QueuedConnection)
                summary.len.add_done_callback(update)
        else:
            self.input = None
        self.__have_new_data = True

    @Inputs.data_subset
    def set_subset_dataset(self, subset: Optional[Table]):
        if subset is not None and not isinstance(subset, SqlTable):
            ids = set(subset.ids)
        else:
            ids = None
        self._subset_ids = ids
        self.__have_new_subset = True

    @Inputs.input_autoshow
    def set_input_autoshow(self, le_str):
        if le_str is not None:
            self.str_WidgetPositionning = str(le_str)

    def handleNewSignals(self):
        super().handleNewSignals()
        self.Warning.non_sortable_input.clear()
        self.Warning.missing_sort_columns.clear()
        data: Optional[Table] = self.input.table if self.input else None
        model = self.input.model if self.input else None

        if self.__have_new_data:
            self._setup_table_view()
            self._update_input_summary()

            if data is not None and self.__pending_sort is not None:
                self.__restore_sort()

            if data is not None and self.__pending_selection is not None:
                selection = self.__pending_selection
                self.__pending_selection = None
                self.set_selection(selection["rows"], selection["columns"])

        if self.__have_new_subset and model is not None:
            model.setSubsetRowIds(self._subset_ids or set())
            self.__have_new_subset = False

        self._setup_view_delegate()

        if self.__have_new_data:
            if self.auto_commit:
                self.commit.now()
            else:
                self.commit.deferred()
            self.__have_new_data = False

    def _setup_table_view(self):
        if self.input is None:
            self.view.setModel(None)
            return

        datamodel = self.input.model
        datamodel.setSubsetRowIds(self._subset_ids or set())

        self._reorder_proxy.setSourceModel(datamodel)
        if self.stored_row_order:
            self._reorder_proxy.setRowOrder(self.stored_row_order)

        view = self.view
        data = self.input.table
        rowcount = data.approx_len()
        view.setModel(self._reorder_proxy)

        vheader = view.verticalHeader()
        if isinstance(vheader, DraggableVerticalHeader):
            vheader.installDragFilter(self._on_row_moved)

        vheader.setDefaultSectionSize(
            view.style().sizeFromContents(
                QStyle.CT_ItemViewItem, view.viewOptions(), QSize(20, 20), view
            ).height() + 2
        )
        vheader.setMinimumSectionSize(5)
        vheader.setSectionResizeMode(QHeaderView.Fixed)

        maxrows = (2 ** 31 - 1) // (vheader.defaultSectionSize() + 2)
        if rowcount > maxrows:
            sliceproxy = TableSliceProxy(parent=view, rowSlice=slice(0, maxrows))
            sliceproxy.setSourceModel(self._reorder_proxy)
            view.setModel(None)
            view.setModel(sliceproxy)

        assert view.model().rowCount() <= maxrows
        assert vheader.sectionSize(0) > 1 or datamodel.rowCount() == 0

        self._setup_view_delegate()
        self._update_variable_labels()

    def _on_row_moved(self, from_pos: int, to_pos: int):
        self._reorder_proxy.moveRow(from_pos, to_pos)
        self.stored_row_order = self._reorder_proxy.getRowOrder()
        self.commit.deferred()

    def _reset_row_order(self):
        if self.input is None:
            return
        n = self.input.model.rowCount()
        self._reorder_proxy.setRowOrder(list(range(n)))
        self.stored_row_order = []
        self.commit.deferred()

    def _update_input_summary(self):
        def format_summary(summary):
            if isinstance(summary, tsummary.ApproxSummary):
                return summary.len.result() if summary.len.done() else summary.approx_len
            return summary.len

        summary, details = self.info.NoInput, ""
        if self.input:
            summary = format_summary(self.input.summary)
            details = format_summary_details(self.input.table)
        self.info.set_input_summary(summary, details)

        if self.input is None:
            self.info_text.setText("No data.")
        else:
            self.info_text.setText("\n".join(tsummary.format_summary(self.input.summary)))

    def _update_variable_labels(self):
        if self.input is None:
            return
        model = self.input.model
        if self.show_attribute_labels:
            model.setRichHeaderFlags(RichTableModel.Labels | RichTableModel.Name)
        else:
            model.setRichHeaderFlags(RichTableModel.Name)

    def _on_distribution_color_changed(self):
        if self.input is None:
            return
        self._setup_view_delegate()

    def _setup_view_delegate(self):
        if self.input is None:
            return
        model = self.input.model
        data = model.source
        class_var = data.domain.class_var
        if self.color_by_class and class_var and class_var.is_discrete:
            color_schema = [QColor(*c) for c in class_var.colors]
        else:
            color_schema = None
        if self.show_distributions:
            delegate = TableBarItemDelegate(
                self.view, color=self.dist_color, color_schema=color_schema
            )
        else:
            delegate = SubsetTableDataDelegate(self.view)
        delegate.subset_opacity = 0.5 if self._subset_ids is not None else 1.0
        self.view.setItemDelegate(delegate)

    def _on_select_rows_changed(self):
        if self.input is None:
            return
        selection_model = self.view.selectionModel()
        selection_model.setSelectBlocks(not self.select_rows)
        if self.select_rows:
            self.view.setSelectionBehavior(QTableView.SelectRows)
            selection_model.select(
                selection_model.selection(),
                QItemSelectionModel.Select | QItemSelectionModel.Rows
            )
        else:
            self.view.setSelectionBehavior(QTableView.SelectItems)

    def restore_order(self):
        """
        Restaure l'ordre original en recréant un _TableModel vierge
        depuis la Table Orange originale (non triée).
        C'est la seule façon fiable de réinitialiser RichTableModel
        qui maintient son propre état de tri interne.
        """
        if self.input is None:
            return

        # Recrée un modèle source vierge depuis la Table originale
        new_model = _TableModel(self.input.table)
        new_model.setSubsetRowIds(self._subset_ids or set())
        self.input = InputData(
            table=self.input.table,
            summary=self.input.summary,
            model=new_model,
        )

        # Rebrancher le proxy sur le nouveau modèle vierge
        self._reorder_proxy.setSourceModel(new_model)

        # Mettre à jour l'affichage
        if self.show_attribute_labels:
            new_model.setRichHeaderFlags(RichTableModel.Labels | RichTableModel.Name)
        else:
            new_model.setRichHeaderFlags(RichTableModel.Name)

        # Réinstaller le drag filter sur le nouveau modèle
        vheader = self.view.verticalHeader()
        if isinstance(vheader, DraggableVerticalHeader):
            vheader.installDragFilter(self._on_row_moved)

        # Remettre l'indicateur de tri à zéro visuellement
        with disconnected(
            self.view.horizontalHeader().sortIndicatorChanged,
            self._on_sort_indicator_changed,
            Qt.UniqueConnection
        ):
            self.view.horizontalHeader().setSortIndicator(-1, Qt.AscendingOrder)

        self.stored_sort = []
        self.stored_row_order = []
        self.Warning.missing_sort_columns.clear()
        self.commit.deferred()

    @Slot()
    def _update_info(self):
        self._update_input_summary()

    def _on_sort_indicator_changed(self, index: int, order: Qt.SortOrder) -> None:
        if index == -1:
            self.stored_sort = []
        elif self.input is not None:
            model = self.input.model
            coldesc = model.columns[index]
            colid = self.__encode_column_id(coldesc)
            order = -1 if order == Qt.DescendingOrder else 1
            self.stored_sort = [(n, d) for n, d in self.stored_sort if n != colid]
            self.stored_sort.append((colid, order))
        self.update_selection()
        self.Warning.missing_sort_columns.clear()

    def set_sort_columns(self, sorting: List[Tuple[str, int]]):
        if self.input is None:
            return
        self.stored_sort = []
        columns = {id_: i for i, id_ in enumerate(self.__header_ids())}
        with disconnected(self.view.horizontalHeader().sortIndicatorChanged,
                          self._on_sort_indicator_changed, Qt.UniqueConnection):
            for colid, order in sorting:
                if colid in columns:
                    self.view.sortByColumn(
                        columns[colid],
                        Qt.AscendingOrder if order == 1 else Qt.DescendingOrder
                    )
                self.stored_sort.append((colid, order))

    def __restore_sort(self) -> None:
        assert self.input is not None
        sort = self.__pending_sort
        self.__pending_sort = None
        if sort is None:
            return
        if not self.view.isSortingEnabled() and sort:
            self.Warning.non_sortable_input()
            self.Warning.missing_sort_columns.clear()
            return
        columns = {id_: i for i, id_ in enumerate(self.__header_ids())}
        missing_columns = []
        sort_ = []
        for colid, order in sort:
            if colid in columns:
                sort_.append((colid, order))
            else:
                missing_columns.append(self.__decode_column_id(colid))
        self.set_sort_columns(sort_)
        if missing_columns:
            self.Warning.missing_sort_columns(", ".join(missing_columns))

    @staticmethod
    def __encode_column_id(coldesc):
        def escape(s):
            return ("\\" + s) if s.startswith("\\") else s
        if isinstance(coldesc, TableModel.Column):
            return escape(coldesc.var.name)
        lookup = ("TARGET", "META", "FEATURES",)
        return f"\\BASKET({lookup[coldesc.role]})"

    @staticmethod
    def __decode_column_id(cid: str) -> str:
        return cid[1:] if cid.startswith("\\") else cid

    def __header_ids(self) -> List[str]:
        if self.input is None:
            return []
        return [self.__encode_column_id(c) for c in self.input.model.columns]

    @staticmethod
    def _normalize_indices(values):
        if values is None:
            return []

        arr = np.asarray(values)

        if arr.ndim == 0:
            return [int(arr.item())]

        return [int(v) for v in arr.ravel().tolist()]

    def update_selection(self, *_):
        self.commit.deferred()

    def set_selection(self, rows: Sequence[int], columns: Sequence[int]) -> None:
        rows = self._normalize_indices(rows)
        columns = self._normalize_indices(columns)
        self.view.setBlockSelection(rows, columns)

    def get_selection(self):
        rows, cols = self.view.blockSelection()
        # blockSelection() peut retourner des index décalés quand un proxy
        # est intercalé. On prend les rows depuis selectedRows() à la place.
        sel_model = self.view.selectionModel()
        correct_rows = [int(i.row()) for i in sel_model.selectedRows()]
        if correct_rows:
            return correct_rows, self._normalize_indices(cols)
        return self._normalize_indices(rows), self._normalize_indices(cols)

    @gui.deferred
    def commit(self):
        selected_data = table = None
        annotated_rows = []

        if self.input is not None:
            model = self.input.model
            table = self.input.table

            if isinstance(table, SqlTable):
                self.Outputs.selected_data.send(selected_data)
                self.Outputs.annotated_data.send(None)
                return

            rowsel, colsel = self.get_selection()
            self.stored_selection = {"rows": list(rowsel), "columns": list(colsel)}

            domain = table.domain
            if len(colsel) < len(domain.variables) + len(domain.metas):
                allvars = domain.class_vars + domain.metas + domain.attributes
                columns = [(c, model.headerData(c, Qt.Horizontal, TableModel.DomainRole))
                           for c in colsel]
                assert all(role is not None for _, role in columns)

                def select_vars(role):
                    return [allvars[c] for c, r in columns if r == role]

                attrs = select_vars(TableModel.Attribute)
                if attrs and issparse(table.X):
                    attrs = table.domain.attributes
                class_vars = select_vars(TableModel.ClassVar)
                metas = select_vars(TableModel.Meta)
                domain = Orange.data.Domain(attrs, class_vars, metas)

            if rowsel:
                row_order = self._reorder_proxy.getRowOrder()
                proxy_rows = [row_order[r] for r in rowsel if r < len(row_order)]
                src_rows = [int(model.mapToSourceRows(r)) for r in proxy_rows]
                selected_data = table.from_table(domain, table, src_rows)
                annotated_rows = src_rows
            else:
                row_order = self._reorder_proxy.getRowOrder()
                sortsection = self.view.horizontalHeader().sortIndicatorSection()

                if sortsection != -1:
                    src_rows = [int(model.mapToSourceRows(r)) for r in row_order]
                    selected_data = table.from_table(table.domain, table, src_rows)
                else:
                    if row_order and row_order != list(range(len(row_order))):
                        src_rows = [int(model.mapToSourceRows(r)) for r in row_order]
                        selected_data = table.from_table(table.domain, table, src_rows)
                    else:
                        selected_data = table

        self.Outputs.selected_data.send(selected_data)
        self.Outputs.annotated_data.send(
            create_annotated_table(table, annotated_rows) if table is not None else None
        )

    def copy(self):
        if self.input is not None:
            mime = table_selection_to_mime_data(self.view)
            QApplication.clipboard().setMimeData(mime, QClipboard.Clipboard)

    def send_report(self):
        if self.input is None:
            return
        model = self.input.model
        self.report_data_brief(model.source)
        self.report_table(self.view)


if __name__ == "__main__":  # pragma: no cover
    WidgetPreview(OWTable).run(
        input_data=Table("iris"),
    )



import os
import sys
import math
import Orange.data
from Orange.widgets import widget
from Orange.widgets.widget import Input, Output
from Orange.data import Table, Domain, StringVariable
from Orange.statistics.basic_stats import BasicStats

if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
    from Orange.widgets.orangecontrib.AAIT.utils.import_uic import uic
    from Orange.widgets.orangecontrib.AAIT.utils.initialize_from_ini import apply_modification_from_python_file
else:
    from orangecontrib.AAIT.utils.import_uic import uic
    from orangecontrib.AAIT.utils.initialize_from_ini import apply_modification_from_python_file


@apply_modification_from_python_file(filepath_original_widget=__file__)
class OWRuleGenerator(widget.OWWidget):
    name = "Rule Generator"
    description = "Génère une règle de filtrage numérique (min <= col <= max) à partir d'une table."
    icon = "icons/rule_generator.svg"
    category = "AAIT - ALGORITHM"
    if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
        icon = "icons_dev/rule_generator.svg"
    priority = 1150
    keywords = "Rule Generator min max filtre bornes"
    gui = os.path.join(os.path.dirname(os.path.abspath(__file__)), "designer/ow_rule_generator.ui")
    want_control_area = False

    class Inputs:
        data = Input("Data", Orange.data.Table)

    class Outputs:
        rules = Output("Rules", Orange.data.Table)

    @Inputs.data
    def set_data(self, data):
        self.data = data
        self.Outputs.rules.send(None)
        self.run()

    def __init__(self):
        super().__init__()
        # Set the fixed width and height of the widget
        self.setFixedWidth(550)
        self.setFixedHeight(300)

        # Load the user interface file
        uic.loadUi(self.gui, self)

        self.data = None
        self.post_initialized()

    def post_initialized(self):
        """Utilisé pour la surcharge uniquement."""
        return

    def run(self):
        self.error("")
        self.warning("")

        if self.data is None:
            self.Outputs.rules.send(None)
            return

        num_attrs = [a for a in self.data.domain.attributes if a.is_continuous]
        if not num_attrs:
            self.warning("Aucune colonne numérique trouvée dans les données.")
            self.Outputs.rules.send(None)
            return

        parts = []
        for attr in num_attrs:                      # <-- fix : BasicStats appelé une variable à la fois
            stat = BasicStats(self.data, attr)
            lo = stat.min
            hi = stat.max
            if math.isnan(lo) or math.isnan(hi):
                continue
            parts.append(f"{lo} <= {attr.name} <= {hi}")

        if not parts:
            self.warning("Toutes les colonnes sont vides.")
            self.Outputs.rules.send(None)
            return

        rule_str = " and ".join(parts)

        sv = StringVariable("regle")
        domain = Domain([], metas=[sv])
        out_table = Table.from_list(domain, [[rule_str]])
        self.Outputs.rules.send(out_table)


if __name__ == "__main__":
    from AnyQt.QtWidgets import QApplication
    app = QApplication(sys.argv)
    my_widget = OWRuleGenerator()
    my_widget.show()
    if hasattr(app, "exec"):
        app.exec()
    else:
        app.exec_()



import os
import sys

import Orange.data
from AnyQt.QtWidgets import QApplication, QLabel
from AnyQt import QtGui
from Orange.widgets import widget

from Orange.widgets.utils.signals import Input, Output
from Orange.widgets.settings import Setting
from AnyQt.QtWidgets import QLineEdit, QTextBrowser, QSpinBox, QDoubleSpinBox
from AnyQt.QtCore import QTimer
from PyQt6.QtWidgets import QCheckBox

if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
    from Orange.widgets.orangecontrib.AAIT.llm import answers_llama
    from Orange.widgets.orangecontrib.AAIT.utils import thread_management,MetManagement, help_management
    from Orange.widgets.orangecontrib.AAIT.utils.import_uic import uic
    from Orange.widgets.orangecontrib.AAIT.utils.initialize_from_ini import apply_modification_from_python_file
else:
    from orangecontrib.AAIT.llm import answers_llama
    from orangecontrib.AAIT.utils import thread_management,MetManagement, help_management
    from orangecontrib.AAIT.utils.import_uic import uic
    from orangecontrib.AAIT.utils.initialize_from_ini import apply_modification_from_python_file


@apply_modification_from_python_file(filepath_original_widget=__file__)
class OWQEdgeLLM(widget.OWWidget):
    name = "Engine LLM"
    description = "Generate a response to a column 'prompt' with a LLM"
    category = "AAIT - LLM INTEGRATION"
    icon = "icons/edge_llm.svg"
    if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
        icon = "icons_dev/edge_llm.svg"
    gui = os.path.join(os.path.dirname(os.path.abspath(__file__)), "designer/owedgellm.ui")
    want_control_area = True
    priority = 1089

    class Inputs:
        data = Input("Data", Orange.data.Table)
        model_path = Input("Model", str, auto_summary=False)

    class Outputs:
        data = Output("Data", Orange.data.Table)

    n_ctx: str = Setting("32768")
    workflow_id = Setting("")

    max_tokens = Setting(0)
    temperature = Setting(0.4)
    top_p = Setting(0.4)
    top_k = Setting(40)
    repeat_penalty = Setting(1.15)
    k_cache = Setting(1)
    v_cache = Setting(1)


    @Inputs.data
    def set_data(self, in_data):
        self.data = in_data
        # if self.autorun:
        #     self.run()

    @Inputs.model_path
    def set_model_path(self, in_model_path):
        self.model_path = in_model_path
        # if self.autorun:
        #     self.run()

    def handleNewSignals(self):
        self.run()

    def __init__(self):
        super().__init__()
        # Qt Management
        self.setFixedWidth(700)
        self.setFixedHeight(700)
        self.mode_convers=False
        uic.loadUi(self.gui, self)
        self.label_description = self.findChild(QLabel, 'Description')
        # Context size
        self.edit_nCtx = self.findChild(QLineEdit, 'lineEdit')
        self.edit_nCtx.setText(str(self.n_ctx))
        self.edit_nCtx.editingFinished.connect(self.update_n_ctx)
        # Context cache
        self.comboKCache.setCurrentIndex(self.k_cache)
        self.comboVCache.setCurrentIndex(self.v_cache)
        self.comboKCache.currentIndexChanged.connect(self.update_cache)
        self.comboVCache.currentIndexChanged.connect(self.update_cache)
        # Workflow ID for streaming
        self.edit_ID = self.findChild(QLineEdit, 'lineEdit_2')
        self.edit_ID.setText(self.workflow_id)
        self.edit_ID.editingFinished.connect(self.update_workflow_id)
        # Bind all spinboxes
        self.edit_max_tokens = self.bind_spinbox("boxMaxTokens", self.max_tokens)
        self.edit_temperature = self.bind_spinbox("boxTemperature", self.temperature, is_double=True)
        self.edit_top_p = self.bind_spinbox("boxTopP", self.top_p, is_double=True)
        self.edit_top_k = self.bind_spinbox("boxTopK", self.top_k)
        self.edit_repeat_penalty = self.bind_spinbox("boxRepeatPenalty", self.repeat_penalty, is_double=True)


        # Text browser
        self.textBrowser = self.findChild(QTextBrowser, 'textBrowser')

        # # Data Management
        self.data = None
        self.model_path = None
        self.thread = None
        self.autorun = True
        self.use_gpu = True
        self.can_run = True
        self.result = None
        self.n_ctx = self.edit_nCtx.text() if self.edit_nCtx.text().isdigit() else "32768"
        self.workflow_id = self.edit_ID.text()

        # Custom updates
        self.post_initialized()
        QTimer.singleShot(0, lambda: help_management.override_help_action(self))

    def update_n_ctx(self):
        value = self.edit_nCtx.text()  # Read the current value
        self.n_ctx = value if value.isdigit() else "32768"  # Default or parsed value
        self.edit_nCtx.setText(self.n_ctx)
       #answers.check_gpu(self.model_path, self) remettre!!!!
        #self.run()

    def update_cache(self):
        self.k_cache = self.comboKCache.currentIndex()
        self.v_cache = self.comboVCache.currentIndex()

    def update_workflow_id(self):
        self.workflow_id = self.edit_ID.text()



    def bind_spinbox(self, name, value, is_double=False):
        widget_type = QDoubleSpinBox if is_double else QSpinBox
        box = self.findChild(widget_type, name)
        box.setValue(value)
        box.editingFinished.connect(self.update_parameters)
        return box

    def update_parameters(self):
        self.max_tokens = self.edit_max_tokens.value()
        self.temperature = self.edit_temperature.value()
        self.top_p = self.edit_top_p.value()
        self.top_k = self.edit_top_k.value()
        self.repeat_penalty = self.edit_repeat_penalty.value()
        print("Updated parameters:")
        print(f"  Max tokens     : {self.max_tokens}")
        print(f"  Temperature    : {self.temperature}")
        print(f"  Top P          : {self.top_p}")
        print(f"  Top K          : {self.top_k}")
        print(f"  Repeat penalty : {self.repeat_penalty}")


    def run(self):
        # Clear error & warning
        self.warning("")
        self.error("")

        # If Thread is already running, interrupt it
        if self.thread is not None:
            if self.thread.isRunning():
                self.thread.safe_quit()

        if self.data is None:
            self.Outputs.data.send(None)
            return

        if self.model_path is None or not os.path.exists(self.model_path):
            self.Outputs.data.send(None)
            return

        domain = self.data.domain

        # Vérifie uniquement les StringVariable
        string_vars = [
            var.name
            for var in list(domain.attributes) + list(domain.class_vars) + list(domain.metas)
            if isinstance(var, Orange.data.StringVariable)
        ]

        has_role = "role" in string_vars
        has_type = "type" in string_vars
        has_content = "content" in string_vars
        has_prompt = "prompt" in string_vars

        # Cas erreur : les 4 colonnes présentes
        if has_role and has_type and has_content and has_prompt:
            self.error("Cannot have role/type/content and prompt simultaneously.")

        # Mode conversation
        elif has_role and has_type and has_content:
            self.mode_convers = True

        # Mode prompt
        elif has_prompt:
            self.mode_convers = False

        # Aucun cas valide
        else:
            self.error(
                "The table must contain either (StringVariable):\n"
                "- role + type + content\n"
                "or:\n"
                "- prompt (image paths)"
            )


        # Différencier les modes Conv / Request ici
        if self.mode_convers:
            if not "role" in self.data.domain:
                self.error('[Conversation enabled] You need a "role" column in your input data.')
                self.Outputs.data.send(None)
                return
            if not "type" in self.data.domain:
                self.error('[Conversation enabled] You need a "type" column in your input data.')
                self.Outputs.data.send(None)
                return
            if not "content" in self.data.domain:
                self.error('[Conversation enabled] You need a "content" column in your input data.')
                self.Outputs.data.send(None)
                return
        else:
            if not "prompt" in self.data.domain:
                self.Outputs.data.send(None)
                return
            if "Answer" in self.data.domain:
                self.error('You cannot have "Answer" in your input data. Please rename or remove the column.')
                self.Outputs.data.send(None)
                return

        #answers_llama.check_gpu(self.model_path, self)
        if not self.can_run:
            return

        query_parameters = {
            "max_tokens": self.max_tokens,
            "temperature": self.temperature,
            "top_p": self.top_p,
            "top_k": self.top_k,
            "repeat_penalty": self.repeat_penalty,
            "k_cache": self.k_cache,
            "v_cache": self.v_cache
        }
        chemin_dossier = MetManagement.get_api_local_folder(workflow_id=self.workflow_id)
        if os.path.exists(chemin_dossier):
            MetManagement.write_file_time(chemin_dossier + "time.txt")

        # Start progress bar
        self.progressBarInit()
        self.textBrowser.setText("")

        # Connect and start thread : main function, progress, result and finish
        # --> progress is used in the main function to track progress (with a callback)
        # --> result is used to collect the result from main function
        # --> finish is just an empty signal to indicate that the thread is finished
        if self.mode_convers:
            self.thread = thread_management.Thread(answers_llama.continue_conversation, self.data, self.model_path,
                                                   self.use_gpu, int(self.n_ctx), query_parameters, self.workflow_id)
        else:
            self.thread = thread_management.Thread(answers_llama.generate_answers, self.data, self.model_path,
                                                   self.use_gpu, int(self.n_ctx), query_parameters, self.workflow_id)
        self.thread.progress.connect(self.handle_progress)
        self.thread.result.connect(self.handle_result)
        self.thread.finish.connect(self.handle_finish)
        self.thread.start()

    def handle_progress(self, progress) -> None:
        action = progress[0]
        value = progress[1]

        if action == "progressBar":
            self.progressBarSet(value)
            # Using the anchor method here too
            self.append_text_at_end("\n\n\n\n")

        elif action == "assistant":
            if value is not None:
                self.append_text_at_end(value)

        elif action == "warning":
            self.warning(value)
        elif action == "error":
            self.error(value)

    def append_text_at_end(self, text):
        # 1. Create a cursor anchored to the document (not the UI focus)
        cursor = QtGui.QTextCursor(self.textBrowser.document())
        # 2. Move this "invisible" cursor to the end
        cursor.movePosition(QtGui.QTextCursor.End)
        # 3. Insert the text : this modifies the underlying data without affecting the scrollbar or the user's cursor
        cursor.insertText(text)

    def handle_result(self, result):
        if result is None:
            self.error("unable to load model")
            self.Outputs.data.send(None)
            return
        try:
            self.result = result
            self.Outputs.data.send(result)
        except Exception as e:
            print("An error occurred when sending out_data:", e)
            self.Outputs.data.send(None)
            return

    def handle_finish(self):
        print("Generation finished")
        self.progressBarFinished()

    def post_initialized(self):
        pass




if __name__ == "__main__":
    app = QApplication(sys.argv)
    my_widget = OWQEdgeLLM()
    my_widget.show()
    if hasattr(app, "exec"):
        app.exec()
    else:
        app.exec_()



import fitz  # PyMuPDF
import os
import re
import json
import copy
import difflib
from json_repair import repair_json


if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
    from Orange.widgets.orangecontrib.AAIT.llm import prompt_management
    from Orange.widgets.orangecontrib.AAIT.llm.answers_llama import run_Qwen3VL_query, run_query, split_think
else:
    from orangecontrib.AAIT.llm import prompt_management
    from orangecontrib.AAIT.llm.answers_llama import run_Qwen3VL_query, run_query, split_think


# Step 1 - Extract PDF pages & convert them to images
def extract_pdf_pages(pdf_path, output_dir):
    image_paths = pdf_to_images(pdf_path, output_dir)
    text_per_page = extract_text_per_page(pdf_path)
    return image_paths, text_per_page


# Step 2 - Find the start of the document
def find_start_of_document(image_paths, model, N=15):
    first_images = image_paths[:N]
    prompt = prompts["find_start_of_document"]
    answer = run_Qwen3VL_query(query=prompt, image_paths=first_images, image_prompts=[], model=model)
    data = answer_to_json(answer)
    start = data["Start_of_Document"]
    if start == "NA":
        start = 0
    else:
        start = int(start) - 1
    return start


# Step 3 - Go through pages to identify the structure of the document
def find_document_structure(image_paths, start, model, batch_size=8, output_dir=None, label=None):
    global_structure = ""
    content_images = image_paths[start:]
    for i in range(0, len(content_images), batch_size):
        if label:
            label.setText(f"Step 2/6 - Generating a global structure - pages {i} to {min(i+batch_size, len(content_images))} out of {len(content_images)} ")
        batch = content_images[i:i + batch_size]
        # Extract page numbers from filenames
        page_numbers = [int(re.findall(r"page_(\d+)\.png", path)[0]) for path in batch]
        numbers_iter = iter(page_numbers)
        # Call the VLM to find headers
        prompt = prompts["generate_TOC"]
        answer = run_Qwen3VL_query(query=prompt, image_paths=batch, image_prompts=[], model=model)
        # Replace IMAGE X → Page X
        answer = re.sub(r"IMAGE \d+", make_replace_image(numbers_iter), answer, flags=re.IGNORECASE)
        global_structure += answer + "\n\n"

    # After computing global_structure in the batch loop, save if demanded
    global_structure = global_structure.replace(r"\n\n", r"\n")
    if output_dir:
        with open(os.path.join(output_dir, "global_structure.txt"), "w", encoding="utf-8") as f:
            f.write(global_structure)

    return global_structure


# Step 4 - Clear the TOC from duplicates and artifacts
def clear_toc(global_structure, model, output_dir=None):
    prompt = prompts["clear_TOC"].format(TEXT=global_structure)
    prompt = prompt_management.apply_prompt_template("qwen", user_prompt=prompt, force_non_thinking=True)
    answer = run_query(prompt=prompt, model=model, max_tokens=0)
    clean_toc = split_think(answer)[1]
    clean_toc = re.sub(r"\*\* ?Table des matières ?\*\*", "", clean_toc).strip()

    if output_dir:
        with open(os.path.join(output_dir, "unstructured_toc.txt"), "w", encoding="utf-8") as f:
            f.write(clean_toc)

    return clean_toc


# Step 5 - Transform the text TOC into a JSON formatted TOC
def toc_to_json(clean_toc, model, output_dir=None):
    prompt = prompts["toc_to_json"].format(TEXT=clean_toc)
    prompt = prompt_management.apply_prompt_template("qwen", user_prompt=prompt, force_non_thinking=True)
    answer = run_query(prompt=prompt, model=model, max_tokens=0)
    json_toc_str = split_think(answer)[1]
    json_toc = answer_to_json(json_toc_str)

    if output_dir:
        with open(os.path.join(output_dir, "json_toc.json"), "w", encoding="utf-8") as f:
            json.dump(json_toc, f, ensure_ascii=False, indent=2)

    return json_toc


# Step 6 - Generate PageIndex
def generate_PageIndex(json_toc, text_per_page, output_dir=None):
    page_index = compute_toc_page_intervals(json_toc, text_per_page=text_per_page)

    if output_dir:
        with open(os.path.join(output_dir, "page_index.json"), "w", encoding="utf-8") as f:
            json.dump(page_index, f, ensure_ascii=False, indent=2)

    return page_index


# Step 7 - Generate a brief note summarizing what can be found in the document
def generate_summary(toc, pdf_name, model, output_dir=None):
    prompt = prompts["short_summary"].format(TEXT=toc, NAME=pdf_name)
    prompt = prompt_management.apply_prompt_template("qwen", user_prompt=prompt)
    answer = run_query(prompt=prompt, model=model, max_tokens=0)
    summary = split_think(answer)[1]

    if output_dir:
        with open(os.path.join(output_dir, "summary.txt"), "w", encoding="utf-8") as f:
            f.write(summary)

    return summary

# Step 7 bis - Generate a brief note summarizing what can be found in the document
def generate_summary_VLM(toc, pdf_name, image_path, model, output_dir=None):
    prompt = prompts["short_summary"].format(TEXT=toc, NAME=pdf_name)
    prompt = prompt_management.apply_prompt_template("qwen", user_prompt=prompt)
    # answer = run_query(prompt=prompt, model=model, max_tokens=0)
    answer = run_Qwen3VL_query(query=prompt, image_paths=image_path, image_prompts=[], model=model)
    summary = split_think(answer)[1]

    if output_dir:
        with open(os.path.join(output_dir, "summary.txt"), "w", encoding="utf-8") as f:
            f.write(summary)

    return summary






# Load intermediate results
def load_file(path, default=None):
    """
    Load a JSON or TXT file depending on extension.

    Args:
        path (str): Path to the file
        default: Value to return if file is missing or invalid

    Returns:
        dict / list (for JSON) or str (for TXT)
    """
    if not os.path.exists(path):
        return default

    ext = os.path.splitext(path)[1].lower()

    try:
        with open(path, "r", encoding="utf-8") as f:
            if ext == ".json":
                return json.load(f)
            elif ext == ".txt":
                return f.read()
            else:
                raise ValueError(f"Unsupported file type: {ext}")
    except (json.JSONDecodeError, OSError, ValueError):
        return default


def verify_database(database_index):
    errors = []
    for entry in database_index:
        doc_id = entry["id"]
        summary = entry["summary"]
        if summary == "An error happened !" or summary == "":
            errors.append(doc_id)
    return errors



def pdf_to_images(pdf_path, output_dir, zoom=2):
    """
    Converts each page of a PDF into PNG images.

    Parameters:
    - pdf_path: str, path to the input PDF file
    - output_dir: str, directory where images will be saved
    - zoom: int, scaling factor for resolution (default 2)

    Returns:
    - image_paths: list of str, paths to the generated image files
    """
    # Ensure output directory exists
    os.makedirs(output_dir, exist_ok=True)
    # Open PDF and define zoom matrix (controls resolution)
    doc = fitz.open(pdf_path)
    mat = fitz.Matrix(zoom, zoom)
    # Iterate through pages and save each as an image
    image_paths = []
    for page_num in range(len(doc)):
        page = doc[page_num]
        pix = page.get_pixmap(matrix=mat)
        img_path = os.path.join(output_dir, f"page_{page_num}.png")
        pix.save(img_path)
        image_paths.append(img_path)
    return image_paths


def extract_text_per_page(pdf_path, start_page=0, end_page=None):
    """
    Extracts text from each page of a PDF and returns a dict: {page_number: text}.

    Parameters:
    - pdf_path: str, path to the PDF
    - start_page: int, 0-based index to start extracting (default 0)
    - end_page: int, 0-based index to end extracting (exclusive, default None = till last page)

    Returns:
    - page_texts: dict mapping 1-based page number -> text
    """
    doc = fitz.open(pdf_path)
    page_texts = {}
    last_page = end_page if end_page is not None else len(doc)
    for i in range(start_page, last_page):
        page = doc[i]
        text = page.get_text("text")
        page_texts[i] = text

    return page_texts


def answer_to_json(answer):
    """
    Extracts and parses a JSON object from a markdown code block.

    Parameters:
    - answer: str, text containing a ```json ... ``` block

    Returns:
    - data: dict, parsed JSON content

    Raises:
    - ValueError: if no valid JSON block is found or parsing fails
    """
    # Find JSON code blocks
    json_matches = re.findall(r"```json([\s\S]*?)```", answer)
    # Expect exactly one JSON block
    if len(json_matches) != 1:
        raise ValueError(f"Wrong format returned by the VLM: {answer}")
    json_str = json_matches[0]
    try:
        # Attempt to repair and parse JSON
        data = json.loads(repair_json(json_str))
    except Exception:
        raise ValueError(f"JSON returned by the VLM had wrong format: {json_str}")
    return data


def make_replace_image(numbers_iter):
    # returns a function compatible with re.sub
    def replace(match):
        return f"Page {next(numbers_iter)}"

    return replace


def compute_toc_page_intervals(json_toc, text_per_page):
    """
    Compute page intervals for each TOC entry.

    Logic:
    - Deepest-level headers: interval = start page → next entry's page - 1 (any level)
    - Higher-level headers:  interval = start page → next sibling/ancestor's page - 1
                             (next entry whose level <= current level)
    """
    toc_copy = copy.deepcopy(json_toc)

    # Normalize level and page to int
    for entry in toc_copy:
        level = entry.get("level", 1)
        entry["level"] = int(level) if str(level).isdigit() else 1
        page = entry.get("page", 0)
        entry["page"] = int(page) if str(page).isdigit() else 0

    n_entries = len(toc_copy)
    max_level = max(entry["level"] for entry in toc_copy)
    last_page = len(text_per_page) - 1

    for i, entry in enumerate(toc_copy):
        level = entry["level"]
        page_start = entry["page"]

        if level == max_level:
            page_end = toc_copy[i + 1]["page"] if i + 1 < n_entries else last_page
        else:
            page_end = last_page
            for j in range(i + 1, n_entries):
                if toc_copy[j]["level"] <= level:
                    page_end = toc_copy[j]["page"]
                    break

        entry["pages"] = f"{page_start} - {page_end}"
        entry.pop("page", None)
        entry.pop("level_identifiable", None)

    return toc_copy





def identify_tool_call(answer):
    """
    Parse a model answer to identify if a tool should be called.

    Returns:
        tool_type: "select_document", "get_pages", or None
        args: tuple of arguments (strings or ints)
    """
    if "```tool" not in answer:
        return None, None
    tool_call = answer.split("```tool")[1].split("```")[0].strip()

    # TOOL 1: view_documents()
    match_doc = re.match(r"view_documents\((.*?)\)", tool_call)
    if match_doc:
        return "view_documents", 0

    # TOOL 2: select_document("doc_id")
    match_doc = re.match(r"select_document\((.*?)\)", tool_call)
    if match_doc:
        doc_name = match_doc.group(1).strip().strip('"').strip("'")
        return "select_document", (doc_name,)

    # TOOL 3: get_pages(start, end)
    match_pages = re.match(r"get_pages\((\d+),\s*(\d+)\)", tool_call)
    if match_pages:
        start, end = int(match_pages.group(1)), int(match_pages.group(2))
        return "get_pages", (start, end)

    # Unknown tool
    return "Unknown tool", None


# TOOL
def view_documents(database_index):
    documents_summaries = []
    for entry in database_index:
        documents_summaries.append({"name": entry["name"], "summary": entry["summary"]})
    document_summaries = f"{documents_summaries}"
    return document_summaries


# TOOL
def select_document(name, database_index):
    """
    Select a document by name. If exact match not found, return the most similar name.

    Parameters:
        name (str): the document name to search
        database_index (list[dict]): list of document entries with "name" keys

    Returns:
        dict: matched document entry (exact or closest)
    """
    # Try exact match first
    for entry in database_index:
        if entry["name"] == name:
            return entry

    # No exact match → find closest
    all_names = [entry["name"] for entry in database_index]
    close_matches = difflib.get_close_matches(name, all_names, n=1, cutoff=0.5)  # cutoff = similarity threshold

    if close_matches:
        # Return entry with the closest name
        closest_name = close_matches[0]
        for entry in database_index:
            if entry["name"] == closest_name:
                return entry

    # If nothing found
    return {}


# TOOL
def get_pages(pdf_path, start=0, end=None, limit=15):
    """
    Extract text from a PDF for a given page range and return as a single string,
    with page headers, missing page notes, separators, and a page limit.

    Parameters:
        pdf_path (str): path to the PDF file
        start (int): 0-based start page (inclusive)
        end (int): 0-based end page (inclusive). If None, goes to last page.
        limit (int): maximum number of pages to actually read

    Returns:
        str: concatenated text with annotations
    """
    # Extract per-page text
    pages_text = extract_text_per_page(pdf_path, start_page=start, end_page=(end + 1 if end is not None else None))

    # Determine actual end page
    last_page = end if end is not None else max(pages_text.keys())

    # Compute how many pages to read according to limit
    num_pages = last_page - start + 1
    read_pages = min(num_pages, limit)
    pages_combined = []

    for i, p in enumerate(range(start, last_page + 1)):
        if i >= read_pages:
            break  # stop reading if limit reached
        if p in pages_text:
            pages_combined.append(f"### Page {p}\n{pages_text[p]}")
        else:
            pages_combined.append(f"### Page {p}\n[Page not available]")

    # If some pages were not read due to limit, add a note
    if num_pages > limit:
        skipped_start = start + read_pages
        skipped_end = last_page
        pages_combined.append(f"### [Pages {skipped_start}-{skipped_end} were not read (limited to {limit} pages at a time). Call get_pages again to access these pages.]")

    return "\n".join(pages_combined)


def clear_ephemeral_messages(ephemeral_messages, conversation, tools=None):
    to_delete = []
    for message in ephemeral_messages:
        if tools is None or message["tool"] in tools:
            idx = message["idx"]
            conversation[idx]["content"] = message["replacement"]
            to_delete.append(message)
    for message in to_delete:
        ephemeral_messages.remove(message)




prompts = {
    "find_start_of_document": """# Contexte
Voici les 15 premières pages d'un document. J'ai besoin de 2 choses :

- S'il y a une Table of Contents, je veux les numéros des images qui en contiennent un morceau.
- Si parmi ces 15 pages, le vrai contenu du document démarre (après la page de garde, les auteurs, les remerciements, etc...), je veux le numéro de l'image où le document démarre.


# Format de réponse
```json
{
	"Table_of_Contents": [x, y, z],
	"Start_of_Document": N
}
```

Si tu n'as pas la réponse, mets une liste vide [] pour table of contents, et "NA" pour start of document.""",


    "generate_TOC": """# Contexte
Tu reçois des images de pages d’un document. Ton objectif est d’identifier uniquement les headers (titres de sections) visibles sur chaque page.

# Définition stricte
Un "header" est un titre structurant le document (section ou sous-section), généralement mis en évidence visuellement.

# Ne sont PAS des headers
- Les définitions, exemples ou notes
- Le texte inline (même en gras ou en majuscules)
- Les légendes, tableaux ou annotations

# Critères pour identifier un header
Un header doit vérifier plusieurs des propriétés suivantes :
- Être visuellement distinct (taille de police plus grande, gras, espacement)
- Être isolé sur une ligne ou en début de section
- Structurer clairement le document (introduction d’une nouvelle partie)

Si tu hésites → IGNORE.

# Instructions
- Pour chaque image, liste uniquement les headers présents
- Recopie les headers EXACTEMENT comme ils apparaissent (ne pas reformuler)
- Une image peut ne contenir aucun header

# Format STRICT de sortie

Image X
- <header exact>
- <header exact>

Si aucun header :
Image X
- Aucun header""",


    # {TEXT}
    "clear_TOC": """# Context
You receive a list of pages containing automatically extracted elements that resemble titles. These elements may be noisy and redundant.

Your goal is to build a **clear, logical, and usable table of contents** for the document.

# Instructions
1. Identify every meaningful topic change, even minor ones.
2. Create sections that reflect the document's outline
3. Each section must span AT MOST 15 pages. Never exceed this limit.
4. Use the extracted titles as hints, but write clean and explicit section names.
5. You must create sub-section if a section is too long (more than 15 pages).
The goal is a readable and practical table of contents for searching the document.

# Output Format (STRICT)
Each header on its own line, with the page where the section begins (only 1 number):  

<header> - <page>  
<header> - <page>  
...

# Data
{TEXT}""",


    # {TEXT}
    "clear_TOC_og": """# Contexte
Voici une extraction page à page d'informations structurelles issues d'un document. L'objectif est de reconstruire la table des matières.

Règles pour les headers répétés :
- Si un header apparaît sur plusieurs pages **consécutives**, ne garder que la première occurrence.
- Si un header apparaît sur des pages **non consécutives**, garder chaque occurrence.

Ignore tout autre artéfact ou information non structurante.


# Format à respecter
Produis une table des matières sous la forme suivante, où Header est la numérotation du header avec son texte, et X est le numéro de page :

** Table des matières**
Header - X
Header - X
etc...


# Extraction page à page
{TEXT}""",


    # $TEXT$
    "toc_to_json": """# Contexte
Voici une table des matières d'un document. Réécris-là dans le format structuré suivant :

```json
[
{{
	"header": "...",
	"level_identifiable": "Oui" ou "Non",
	"level": "...",
	"page": "..."
}}
]
```

Tu devras ajouter l'information "level" pour chacun de ces headers. Le "level" correspond au niveau de hiérarchie du header.
Un "level" est identifiable si le header est accompagné d'une numérotation, comme par exemple "1.1" (level vaut 2) ou "1.3.1" (level vaut 3).
Si le "level" n'est pas identifiable, note 1. 



# Table des matières
{TEXT}""",



    # {TEXT}
    "short_summary": """# Contexte
Tu vas recevoir la table des matières d'un document, ainsi que la page de garde. Génère une **note très concise et directe** décrivant le sujet et ce que l'on peut trouver dans ce document. 
L'objectif est que, en lisant cette note, on puisse rapidement savoir si ce document contient les informations recherchées.


# Instructions :
- Résume le sujet et le contenu : thèmes, concepts, procédures, informations techniques ou idées principales.
- N'utilise pas plus de 3 phrases.
- Utilise des mots-clés pertinents et séparés par des virgules.
- N'ajoute pas d'opinion, commentaire ou phrase explicative.


# Extrait du document {NAME}
{TEXT}""",


    # {PAGE_INDEX}
    # {QUESTION}
    "search_agent": """# Contexte
Tu es un assistant avec accès à deux outils :
1. Un outil pour sélectionner un document
2. Un outil pour parcourir les pages d'un document sélectionné

Tu disposes d’un ensemble de documents, chacun décrit par un résumé.


# Règles
- Tu n'as PAS le droit d'inventer ou de deviner.
- Tu n'as PAS le droit de conclure que la réponse n'existe pas **sans** avoir lu de pages pertinentes avec l’outil.
- Tu peux utiliser les outils autant que tu veux jusqu'à avoir suffisamment d'informations.

---

# Comportement attendu
## Étape 1 — Sélection du document
- Si la question concerne les documents → tu DOIS d'abord sélectionner un document
- Choisis le document le plus pertinent en te basant sur les résumés fournis
- Tu ne peux pas appeler `get_pages` sans avoir sélectionné un document

## Étape 2 — Lecture des pages
- Une fois le document sélectionné → tu peux appeler `get_pages`

---

# Utilisation d'un outil
## Outil 1 - Sélection d'un document
Réponds avec

```tool
select_document(<pdf_name>)
```

puis attend le résultat avant de continuer tout réflexion ou rédaction de réponse. 
Tu vas recevoir un plan du document, te permettant de cibler les pages qui t'intéressent.


## Outil 2 - Lecture de pages
Réponds avec

```tool
get_pages(<page_debut>, <page_fin>)
```

puis attend le résultat avant de continuer toute réflexion ou rédaction de réponse.

---

# Arrêt
Lorsque tu as suffisamment d'informations pour répondre :
- Tu dois répondre directement à la question
- Tu ne dois PAS appeler l'outil
- Ta réponse doit être complète et précise

Format de réponse final :
```final
<ta réponse ici>
```

---

# Documents disponibles
{DOCUMENTS_SUMMARIES}


# Question
{QUESTION}"""
}


import ast
import copy
import os
import re
import numpy as np
try:
    import GPUtil #sometimes errors occurs on gpu testing
except:
    pass
import psutil
import base64
import ntpath
import platform
from llama_cpp import Llama
from jinja2 import Template

from Orange.data import Domain, StringVariable, Table


if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
    from Orange.widgets.orangecontrib.AAIT.llm import prompt_management,handler_llama
    from Orange.widgets.orangecontrib.AAIT.utils import MetManagement
else:
    from orangecontrib.AAIT.llm import prompt_management,handler_llama
    from orangecontrib.AAIT.utils import MetManagement



supported_VLM = {
    "Qwen3-VL-8B-Instruct-Q4_K_M.gguf": "mmproj-F16.gguf",
    "Qwen3.5-9B-Q6_K.gguf": "mmproj-F16.gguf",
    "Qwen3.5-4B-Q4_K_M.gguf": "mmproj-F16.gguf"
}# manque Qwen3-VL



# Information (useless in the code)
format_for_messages = [
    {"role": "system", "content": str},

    {
        "role": "user",
        "content": [
            {"type": "text", "text": str},
            {"type": "image_url", "image_url": {"url": str}}
        ]
    },

    {
        "role": "assistant",
        "content": [
            {"type": "text", "text": str},
            {"type": "image_url", "image_url": {"url": str}}
        ]
    }
]



def check_gpu(model_path, argself, mmproj_path=None, n_ctx=0):
    """
    Checks if the GPU has enough VRAM to load a model.

    Args:
        model_path (str): Path to the model file.
        argself (OWWidget): OWQueryLLM object.

    Returns:
        bool: True if the model can be loaded on the GPU, False otherwise.
    """
    argself.error("")
    argself.warning("")
    argself.information("")
    argself.can_run = True
    argself.use_gpu = True

    # Resolve effective context size
    if not n_ctx:
        n_ctx = getattr(argself, "n_ctx", 4096)

    try:
        n_ctx = int(n_ctx)
    except Exception:
        n_ctx = 4096

    # attention bien faire la suite dans un try exept car gputil peut etre capricieux
    token_weight = 0.13
    if model_path is None:
        argself.use_gpu = False
        return
    if platform.system() != "Windows":
        argself.use_gpu = False
        return
    if not model_path.endswith(".gguf"):
        argself.use_gpu = False
        argself.can_run = False
        argself.error("Model is not compatible. It must be a .gguf format.")
        return
    # Calculate the model size in MB, including mmproj if provided with a 1500 MB buffer
    model_size = os.path.getsize(model_path) / (1024 ** 3) * 1000
    if mmproj_path and os.path.isfile(mmproj_path):
        mmproj_size = os.path.getsize(mmproj_path) / (1024 ** 3) * 1000
        print(f"mmproj size: {mmproj_size/1000:.2f}GB")
        model_size += mmproj_size

    kv_cache_mb = (n_ctx * 150_000) / (1024 ** 2)
    model_size += kv_cache_mb

    print(f"KV cache estimate: {kv_cache_mb/1000:.2f}GB")
    print(f"Required memory total: {model_size/1000:.2f}GB")
    # If there is no GPU, set use_gpu to False
    if len(GPUtil.getGPUs()) == 0:
        argself.use_gpu = False
        argself.information("Running on CPU. No GPU detected.")
        return
    # Else
    else:
        # Get the available VRAM on the first GPU
        gpu = GPUtil.getGPUs()[0]
        free_vram = gpu.memoryFree
        print(f"Free VRAM: {free_vram/1000:.2f}GB")

    # If there is not enough VRAM on GPU
    if free_vram < model_size:
        # Set use_gpu to False
        argself.use_gpu = False
        # Check for available RAM
        available_ram = psutil.virtual_memory().available / 1024 / 1024
        if available_ram < model_size:
            argself.can_run = False
            argself.error(f"Cannot run. Both GPU and CPU are too small for this model (required: {model_size/1000:.2f}GB).")
            return
        else:
            argself.warning(f"Running on CPU. GPU seems to be too small for this model (available: {free_vram/1000:.2f}GB || required: {model_size/1000:.2f}GB).")
            return
    # If there is enough space on GPU
    else:
        try:
            # Load the model and test it
            # model = GPT4All(model_name=model_path, model_path=model_path, n_ctx=int(argself.n_ctx),
            #                 allow_download=False, device="cuda")
            # answer = model.generate("What if ?", max_tokens=3)
            # # If it works, set use_gpu to True
            argself.use_gpu = True
            argself.information("Running on GPU.")
            return
        # If importing Llama and reading the model doesn't work
        except Exception as e:
            # Set use_gpu to False
            argself.use_gpu = False
            argself.warning(f"GPU cannot be used. (detail: {e})")
            return


def count_tokens(model, message, image_token_cost=1968, overhead_size=4):
    """
    Estimate token count for a single message.
    """
    # CASE 1: pure string
    if isinstance(message, str):
        return len(model.tokenize(message.encode("utf-8"))) + overhead_size

    total_tokens = 0
    # CASE 2: list (multimodal)
    if isinstance(message, list):
        for item in message:
            if not isinstance(item, dict):
                continue
            # TEXT
            if "text" in item:
                text = item["text"]
                total_tokens += len(model.tokenize(text.encode("utf-8"))) + overhead_size
            # IMAGE
            elif "image_url" in item:
                total_tokens += image_token_cost + overhead_size
            # Else, weird
            else:
                print("Wrong message format !")
    else:
        print("Wrong message format !")
    return total_tokens


def load_model(model_path, use_gpu, n_ctx=10000, k_cache=0, v_cache=0, verbose=False):
    """
    Charge un modèle GGUF avec llama_cpp.Llama.

    - use_gpu=True : tente d'utiliser l'accélération (Metal/CUDA/Vulkan selon build)
      en mettant n_gpu_layers à -1 (= toutes les couches si possible).
    - use_gpu=False : CPU only (n_gpu_layers=0).
    """
    if not os.path.exists(model_path):
        print(f"Model could not be found: {model_path} does not exist")
        return

    try:
        # n_gpu_layers : -1 = toutes les couches si le binaire a un backend GPU (Metal/CUDA/Vulkan)
        n_gpu_layers = -1 if use_gpu else 0

        # n_threads : par défaut tous les cœurs logiques dispo moins 1 (pour avoir l'interface graphique qui ne freeze pas)
        n_threads = max(1, (os.cpu_count()-1 or 1))

        # NOTE : llama_cpp utilise n_ctx pour la taille de contexte
        model = Llama(
            model_path=model_path,
            n_ctx=n_ctx,
            n_threads=n_threads,
            n_gpu_layers=n_gpu_layers,
            # Quelques réglages sûrs
            use_mmap=True,
            use_mlock=False,
            embedding=False,
            verbose=verbose,
            type_k=k_cache,
            type_v=v_cache
        )
        return model
    except Exception as e:
        print("Failed to load model with llama_cpp:", e)
        return


def generate_answers(table, model_path, use_gpu=False, n_ctx=4096, query_parameters=None, workflow_id="", progress_callback=None, argself=None):
    """
    Identique en signature/comportement, mais utilise llama_cpp sous le capot.
    """
    # Copie des données d'entrée
    data = copy.deepcopy(table)
    attr_dom = list(data.domain.attributes)
    metas_dom = list(data.domain.metas)
    class_dom = list(data.domain.class_vars)

    # Chargement modèle (llama_cpp)
    model = load_model_with_handler(model_path=model_path, use_gpu=use_gpu, n_ctx=n_ctx, verbose=True)
    with_handler = True
    if model is None:
        model = load_model(model_path=model_path,
                           use_gpu=use_gpu,
                           n_ctx=n_ctx,
                           k_cache=query_parameters["k_cache"],
                           v_cache=query_parameters["v_cache"],
                           verbose=True)
        with_handler = False
    if model is None:
        return None

    # Paramètres de génération par défaut
    if query_parameters is None:
        query_parameters = {"max_tokens": 4096, "temperature": 0.4, "top_p": 0.4, "top_k": 40, "repeat_penalty": 1.15}

    # Génération sur la colonne "prompt", fonctionnement ligne à ligne

    rows = []
    for i, row in enumerate(data):
        features = list(data[i])
        metas = list(data.metas[i])
        prompt = row["prompt"].value

        system_prompt = row["system prompt"].value if "system prompt" in data.domain else ""
        assistant_prompt = row["assistant prompt"].value if "assistant prompt" in data.domain else ""

        # Si ce n'est pas un VLM:
        if not with_handler:
            prompt = prompt_management.apply_prompt_template(
                model_path,
                user_prompt=prompt,
                assistant_prompt=assistant_prompt,
                system_prompt=system_prompt
            )

            prompt = handle_context_length(prompt, model, n_ctx, method="truncate", margin=query_parameters["max_tokens"], progress_callback=progress_callback)

            answer = run_query(
                prompt,
                model=model,
                max_tokens=query_parameters["max_tokens"],
                temperature=query_parameters["temperature"],
                top_p=query_parameters["top_p"],
                top_k=query_parameters["top_k"],
                repeat_penalty=query_parameters["repeat_penalty"],
                workflow_id=workflow_id,
                argself=argself,
                progress_callback=progress_callback
            )

        else:
            # Parsing et regroupement des prompts / image path
            image_paths = [p.strip() for p in row["image paths"].value.split(";")] if "image paths" in data.domain else []
            message_rows = [["user", "text", prompt]]
            for image_path in image_paths:
                message_rows.append(["user", "image", image_path])
            if assistant_prompt:
                message_rows.append(["assistant", "text", assistant_prompt])

            # Création d'une table pour utiliser la helper function table_to_messages
            v1 = StringVariable("role")
            v2 = StringVariable("type")
            v3 = StringVariable("content")
            temp_domain = Domain([], metas=[v1, v2, v3])
            temp_table = Table.from_list(temp_domain, rows=message_rows)
            messages = table_to_messages(temp_table)

            answer = chat_completion_with_handler(messages=messages,
                                                  model=model,
                                                  parameters=query_parameters,
                                                  workflow_id=workflow_id,
                                                  progress_callback=progress_callback,
                                                  argself=argself)

        if answer == "":
            answer = (
                "Error: The answer could not be generated. Your prompt might be too long, or the model architecture you tried to use is possibly "
                f"not supported yet.\n\nModel name: {ntpath.basename(model_path)}"
            )

        thinking = ""
        matches = re.findall(r"<think>[\s\S]*?</think>", answer)
        if matches:
            thinking = matches[0]
            answer = answer.replace(thinking, "").strip()
        metas += [answer, thinking]
        rows.append(features + metas)

        if progress_callback is not None:
            progress_value = float(100 * (i + 1) / len(data))
            progress_callback(("progressBar", progress_value))

        if argself is not None and getattr(argself, "stop", False):
            break

    # Ajouter la colonne "Answer" en metas
    answer_dom = [StringVariable("Answer"), StringVariable("Thinking")]

    domain = Domain(attributes=attr_dom, metas=metas_dom + answer_dom, class_vars=class_dom)
    out_data = Table.from_list(domain=domain, rows=rows)
    return out_data


class StopCallback:
    def __init__(self, stop_sequences, widget_thread=None):
        self.stop_sequences = stop_sequences
        self.recent_tokens = ""
        self.returning = True  # Store the last valid token before stopping
        self.widget_thread = widget_thread

    def __call__(self, token_id, token):
        # Stop in case thread is stopped
        if self.widget_thread:
            if self.widget_thread.stop:
                return False

        # Stop in case stop word has been met
        if not self.returning:
            return False
        self.recent_tokens += token

        # Check if any stop sequence appears
        for stop_seq in self.stop_sequences:
            if stop_seq in self.recent_tokens:
                self.returning = False  # Stop the generation, but allow the last token

        return True  # Continue generation


def write_tokens_to_file(token: str, workflow_id=""):
    chemin_dossier = MetManagement.get_api_local_folder(workflow_id=workflow_id)
    if os.path.exists(chemin_dossier):
        MetManagement.write_file_time(chemin_dossier + "time.txt")
        filepath = os.path.join(chemin_dossier, "chat_output.txt")
        with open(filepath, "a", encoding="utf-8") as f:
            f.write(token)
            f.flush()


def run_query(prompt, model, max_tokens=4096, temperature=0.4, top_p=0.8, top_k=50, repeat_penalty=1.15,
              workflow_id="", argself=None, progress_callback=None):
    """
    Version llama_cpp avec streaming.
    On garde la même signature et le même contrat de retour.
    """


    # Séquences d'arrêt à filtrer du résultat final
    stop_sequences = ["<|endoftext|>", "### User", "<|im_end|>", "<|im_start|>", "<|im_end>", "<im_end|>", "<im_end>"]
    callback_instance = StopCallback(stop_sequences, argself)

    # Paramètres de sampling mappés vers llama_cpp
    gen_kwargs = dict(
        max_tokens=max_tokens,
        temperature=temperature,
        top_p=top_p if top_p else 1.0,   # top_p=0 désactive → on met 1.0
        top_k=top_k if top_k else 0,     # top_k=0 désactive
        repeat_penalty=repeat_penalty,
        stream=True,
    )

    answer = ""

    # IMPORTANT :
    # - On utilise create_completion (prompt-style) pour rester compatible avec ton templating actuel.
    # - Le générateur renvoie des chunks contenant choices[0].text.
    try:
        stream = model(prompt=prompt, **gen_kwargs)

        for chunk in stream:
            # Récupérer le texte incrémental
            token = chunk["choices"][0].get("text", "")
            if not token:
                continue

            # Callback d'arrêt custom (on simule token_id=None)
            if not callback_instance(None, token):
                # On stoppe proprement le flux (consommation du générateur non nécessaire)
                answer += token  # on peut inclure le dernier token si souhaité
                break

            answer += token
            write_tokens_to_file(token, workflow_id)
            print(token, end="")

            if progress_callback is not None:
                progress_callback(("assistant", token))

            if argself is not None and getattr(argself, "stop", False):
                # Arrêt demandé de l'extérieur
                return answer

    except Exception as e:
        # En cas d'erreur pendant la génération, on retourne ce qu'on a + log
        print("Generation error (llama_cpp):", e)

    # Nettoyage des séquences d'arrêt
    for stop in stop_sequences:
        if stop:
            answer = answer.replace(stop, "")

    return answer


def split_think(answer: str):
    # Extract think content (if any)
    think_match = re.search(r"(.*?)</think>", answer, flags=re.DOTALL)
    think_text = think_match.group(1).strip() if think_match else ""
    # Remove think block from the final answer
    final_answer = re.sub(r".*?</think>", "", answer, flags=re.DOTALL).strip()
    return think_text, final_answer


def handle_context_length(prompt, model, n_ctx, method="truncate", margin=0, progress_callback=None):
    """
    Truncate a prompt to fit within n_ctx tokens, leaving margin for generation.
    Safely handles edge cases where limit <= 0.
    """
    # Keep a margin for generated tokens
    limit = max(n_ctx - margin, 0)  # clamp to at least 0

    if method == "truncate":
        tokens = model.tokenize(prompt.encode("utf-8"))  # pass string, not bytes
        initial_length = len(tokens)
        if initial_length > limit:
            # take last `limit` tokens safely
            tokens = tokens[-limit:] if limit > 0 else []
            truncated_length = len(tokens)
            prompt = model.detokenize(tokens).decode("utf-8") if tokens else ""
            if progress_callback:
                warning = (
                    f"Complete prompt contains {initial_length} tokens - context limit is {limit} (Context length - Max tokens). "
                    f"The {truncated_length} last tokens have been kept in the prompt."
                )
                progress_callback(("warning", warning))
        return prompt
    elif method == "summarize":
        pass
    else:
        return prompt


def handle_long_messages(messages, model, n_ctx, method="truncate", margin=0, progress_callback=None):
    limit = max(n_ctx - margin, 0)

    if method == "truncate":
        return _handle_truncate(messages, model, limit, progress_callback)
    elif method == "summarize":
        # Placeholder for future implementation
        raise NotImplementedError("Summarization method is not yet implemented.")
    else:
        raise ValueError(f"Unknown method: {method}")


def _handle_truncate(messages, model, limit, progress_callback):
    kept_messages = []
    total_tokens = 0
    system_msg = None

    # 1. Separate and count the system message first
    if messages and messages[0]["role"] == "system":
        system_msg = messages[0]
        text = messages[0]["content"]
        total_tokens += count_tokens(model=model, message=text)

        # If the system message itself exceeds the limit, we're in trouble
        if total_tokens > limit:
            if progress_callback:
                progress_callback(("error", "System message exceeds context limit."))
            return [system_msg]

    # 2. Iterate through the rest of the messages (newest to oldest)
    # We skip index 0 if it's the system message
    chat_history = messages[1:] if system_msg else messages

    for msg in reversed(chat_history):
        content = msg["content"]
        tokens = count_tokens(model=model, message=content)

        if total_tokens + tokens > limit:
            if progress_callback:
                warning = (
                    f"Context limit reached ({limit} tokens). "
                    f"Keeping {len(kept_messages)} recent messages plus system prompt."
                    f"Remember that an image ≈ 2000 tokens."
                )
                progress_callback(("warning", warning))
            break

        kept_messages.append(msg)
        total_tokens += tokens

    # 3. Restore chronological order
    kept_messages.reverse()

    # 4. Re-attach the system message at the very top
    if system_msg:
        kept_messages.insert(0, system_msg)

    return kept_messages



# For pure display
def conversation_to_text(conversation):
    rendered_conversation = ""
    for messages in conversation:
        rendered_conversation += f"# {messages['role']}\n"
        if isinstance(messages["content"], str):
            rendered_conversation += messages["content"]
        else:
            for message in messages["content"]:
                if message["type"] == "text":
                    rendered_conversation += message["text"]
                elif message["type"] == "image_url":
                    rendered_conversation += "[IMAGE]"
        rendered_conversation += "\n"
    return rendered_conversation


def continue_conversation(table, model_path, use_gpu=False, n_ctx=32768, query_parameters=None, workflow_id="", progress_callback=None, argself=None):
    """
    Continues a multimodal conversation from an Orange data table using a local language model.

    This function converts a structured table into chat messages, loads the appropriate model
    (standard LLM or vision-language model), applies token/context handling, and generates a
    response using either a dedicated chat handler or a generic prompt-based pipeline. The
    generated assistant reply is then appended back into the original Orange Table format.

    Parameters:
    ----------
    table : Orange.data.Table
        Input conversation table containing rows with role/type/content metadata.
    model_path : str
        Path to the local model file or directory.
    use_gpu : bool, optional
        Whether to enable GPU acceleration for model inference.
    n_ctx : int, optional
        Maximum context window size for the model.
    query_parameters : dict, optional
        Generation parameters such as:
        - max_tokens
        - temperature
        - top_p
        - top_k
        - repeat_penalty
        Also may include cache settings (k_cache, v_cache).
    workflow_id : str, optional
        Identifier used for tracking or logging the generation workflow.
    progress_callback : callable, optional
        Callback function for reporting generation progress (UI updates, logs, etc.).
    argself : object, optional
        Optional reference to a widget or external context for callbacks.

    Returns:
    -------
    Orange.data.Table
        A new table containing the original conversation plus one additional row
        representing the assistant's generated response.

    Returns None if model loading or message conversion fails.
    """
    # Copie des données d'entrée
    data = copy.deepcopy(table)

    # Chargement modèle (llama_cpp)
    model_name = os.path.basename(model_path)

    if handler_llama.find_mmproj_path(model_path) is not None:
        model = load_model_with_handler(model_path, n_ctx=n_ctx, use_gpu=use_gpu, verbose=True)
        with_handler = True
    else:
        model = load_model(model_path=model_path,
                           use_gpu=use_gpu,
                           n_ctx=n_ctx,
                           k_cache=query_parameters["k_cache"],
                           v_cache=query_parameters["v_cache"])
        with_handler = False
    if model is None:
        return None

    # Default generation parameters
    if query_parameters is None:
        query_parameters = {"max_tokens": 0, "temperature": 0.4, "top_p": 0.4, "top_k": 40, "repeat_penalty": 1.15}

    # Build the conversation from table
    messages = table_to_messages(data)
    if not messages:
        return
    messages = handle_long_messages(messages, model, n_ctx, method="truncate", margin=query_parameters["max_tokens"], progress_callback=progress_callback)

    ### GENERATE ANSWER
    if with_handler:
        answer = chat_completion_with_handler(messages=messages,
                                              model=model,
                                              parameters=query_parameters,
                                              workflow_id=workflow_id,
                                              progress_callback=progress_callback,
                                              argself=argself)
    else:
        try:
            print("Trying native prompt formating...")
            chat_template = model.metadata["tokenizer.chat_template"]
            if not is_multimodal_template(chat_template):
                messages = flatten_multimodal_messages(messages, drop_images=True)
            template = Template(chat_template)
            prompt = template.render(messages=messages, tools=None, add_generation_prompt=True)
            print("Successfully generated prompt.")
        except Exception as e:
            print(f"An error happened: {e}. Falling back to generic prompt formating...")
            prompt = prompt_management.apply_template_to_conversation_2(model.model_path, conversation=messages)
        answer = run_query(
            prompt,
            model=model,
            max_tokens=query_parameters["max_tokens"],
            temperature=query_parameters["temperature"],
            top_p=query_parameters["top_p"],
            top_k=query_parameters["top_k"],
            repeat_penalty=query_parameters["repeat_penalty"],
            workflow_id=workflow_id,
            argself=argself,
            progress_callback=progress_callback
        )
    thinking, answer = split_think(answer)

    # Create output table
    meta = data.metas[-1].copy()
    meta_names = [m.name for m in data.domain.metas]
    meta[meta_names.index("role")] = "assistant"
    meta[meta_names.index("type")] = "text"
    meta[meta_names.index("content")] = answer
    empty_x = np.zeros((1, len(data.domain.attributes)))

    new_row = Table.from_numpy(
        data.domain,
        X=empty_x,
        metas=np.array([meta], dtype=object)
    )
    out_data = Table.concatenate([data, new_row])
    return out_data


# Should replace load_Qwen3VL (more generic)
def load_model_with_handler(model_path, n_ctx=32768, use_gpu=True, verbose=True):
    """
    Loads a multimodal (vision-language) model using a dedicated chat handler and GGUF backend.

    This function initializes a model with optional GPU acceleration and attaches the
    appropriate multimodal projector (mmproj) required for vision capabilities. It currently
    supports specific VLM architectures (e.g., Qwen3-VL) by selecting the correct chat handler
    and configuring the Llama backend accordingly.

    Parameters:
    ----------
    model_path : str
        Path to the GGUF model file.
    n_ctx : int, optional
        Context window size for the model (default is 32768).
    use_gpu : bool, optional
        If True, enables GPU acceleration by offloading layers.
    verbose : bool, optional
        If True, enables detailed logging during model loading and inference.

    Returns:
    -------
    Llama or None
        A configured Llama model instance with an attached multimodal chat handler,
        or None if required projector files are missing or the model is unsupported.
    """
    model_name = os.path.basename(model_path)
    mmproj_path = handler_llama.find_mmproj_path(model_path)

    if mmproj_path is None or not os.path.exists(mmproj_path) :
        print(f"Something is wrong with your model : couldn't find the projector {mmproj_path}")
        return None

    n_gpu_layers = -1 if use_gpu else 0
    model = None
    chat_handler = handler_llama.get_chat_handler(model_path, mmproj_path, verbose=verbose, use_gpu=use_gpu)
    model = Llama(model_path=model_path,
                  chat_handler=chat_handler,
                  n_ctx=n_ctx,
                  n_gpu_layers=n_gpu_layers,
                  verbose=verbose)

    return model


def run_Qwen3VL_query(query, image_paths, image_prompts, model, system_prompt=" ", workflow_id="", progress_callback=None):
    image_messages = []
    for i, image_path in enumerate(image_paths):
        if os.path.exists(image_path):
            data_uri = convert_to_uri(image_path)
            if not data_uri.startswith("data"):
                progress_callback(("error", data_uri))
                return "The image could not be processed"
            # Add prompt first, if available
            if image_prompts and i < len(image_prompts):
                image_messages.append({
                    "type": "text",
                    "text": image_prompts[i]
                })
            # Then add the image
            image_messages.append({
                "type": "image_url",
                "image_url": {"url": data_uri}
            })
    image_messages.append({"type": "text", "text": query})

    messages = [{"role": "system", "content": system_prompt},
                {"role": "user", "content": image_messages}]
    generator = model.create_chat_completion(messages=messages, stream=True)

    full_response = ""
    for chunk in generator:
        # chunk is a dict, often with a 'choices' list
        for choice in chunk.get("choices", []):
            # Each choice may have a 'delta' dict with 'content'
            delta = choice.get("delta", {})
            token = delta.get("content")
            if token:
                full_response += token
                write_tokens_to_file(token, workflow_id)
                if progress_callback is not None:
                    progress_callback(("assistant", token))
    return full_response


# Should replace run_Qwen3VL_query (more generic)
def chat_completion_with_handler(messages, model, parameters, workflow_id="", progress_callback=None, argself=None):
    """
    Generates a streaming chat completion using a multimodal-capable Llama model handler.

    This function sends a list of chat messages to the model and streams the response token
    by token. It aggregates the generated tokens into a full response while optionally
    reporting progress in real time via callbacks and persisting tokens to disk.

    The generation can be interrupted externally via the `argself.stop` flag.

    Parameters:
    ----------
    messages : list
        List of chat messages in OpenAI-style format, including role and multimodal content.
    model : Llama
        Loaded Llama model instance with an attached chat handler.
    parameters : dict
        Generation parameters including:
        - temperature
        - top_p
        - top_k
        - repeat_penalty
        - max_tokens
    workflow_id : str, optional
        Identifier used for logging or tracking streamed tokens.
    progress_callback : callable, optional
        Callback function receiving streamed tokens for UI or live display.
    argself : object, optional
        External controller object that may contain a `.stop` flag to interrupt generation.

    Returns:
    -------
    str
        The full generated assistant response as a single concatenated string.
    """
    thinks = handler_llama.is_a_thinking_model(model)
    think_token_added = False
    print("Messages:", messages)
    print("\n\n")
    print(model)
    generator = model.create_chat_completion(messages=messages,
                                             temperature=parameters["temperature"],
                                             top_p=parameters["top_p"],
                                             top_k=parameters["top_k"],
                                             repeat_penalty=parameters["repeat_penalty"],
                                             max_tokens=parameters["max_tokens"],
                                             stream=True)
    full_response = ""
    for chunk in generator:
        # chunk is a dict, often with a 'choices' list
        for choice in chunk.get("choices", []):
            # Each choice may have a 'delta' dict with 'content'
            delta = choice.get("delta", {})
            token = delta.get("content")

            if thinks and not think_token_added:
                thinking_token = "<think>\n"
                full_response += thinking_token
                write_tokens_to_file(thinking_token)
                if progress_callback is not None:
                    progress_callback(("assistant", thinking_token))
                think_token_added = True

            if token:
                full_response += token
                write_tokens_to_file(token, workflow_id)
                if progress_callback is not None:
                    progress_callback(("assistant", token))
                if argself is not None and getattr(argself, "stop", False):
                    return full_response
    return full_response



def table_to_messages(data):
    """
    Converts a structured table of role/type/content rows into a chat-formatted message list.

    This function groups consecutive USER and ASSISTANT rows into single messages with multimodal
    content (text and images), while keeping SYSTEM messages as standalone entries. It also converts
    image paths into data URIs when valid, enabling multimodal compatibility.

    Parameters:
    ----------
    data : Orange.data.Table
        Iterable of rows where each row contains:
        - "role": role enum (system, user, assistant)
        - "type": content type enum (text or image)
        - "content": actual content value (string path or text)

    Returns:
    -------
    list
        A list of message dictionaries in chat format:
        - {"role": "system", "content": str}
        - {"role": "user"/"assistant", "content": [{"type": "text", ...}, {"type": "image_url", ...}]}

    Returns None if an image conversion error occurs.
    """
    messages = []
    current_message = None
    for row in data:
        role = row["role"].value
        typ = row["type"].value
        content = row["content"].value

        # SYSTEM → always standalone
        if role == "system":
            messages.append({"role": "system", "content": content})
            current_message = None
            continue

        # USER / ASSISTANT → group into one message
        if current_message is None or current_message["role"] != role:
            current_message = {"role": role, "content": []}
            messages.append(current_message)

        # Add content item
        if typ == "text":
            current_message["content"].append({"type": "text", "text": content})
        elif typ == "image":
            content = content.strip("'").strip('"')
            if os.path.exists(content):
                data_uri = convert_to_uri(content)
                if not data_uri.startswith("data"):
                    current_message["content"].append({"type": "text", "text": "Error: image could not be loaded !"})
                else:
                    current_message["content"].append({"type": "image_url", "image_url": {"url": data_uri}})
            else:
                current_message["content"].append({"type": "text", "text": "PathError: the image doesn't exist !"})
    return messages

_IMAGE_MIME_TYPES = {
    # Most common formats
    '.png':  'image/png',
    '.jpg':  'image/jpeg',
    '.jpeg': 'image/jpeg',
    '.gif':  'image/gif',
    '.webp': 'image/webp',
    '.svg':  'image/svg+xml',
    '.svgz': 'image/svg+xml',

    # Next-generation formats
    '.avif': 'image/avif',
    '.heic': 'image/heic',
    '.heif': 'image/heif',
    '.heics': 'image/heic-sequence',
    '.heifs': 'image/heif-sequence',

    # Legacy / Windows formats
    '.bmp':  'image/bmp',
    '.dib':  'image/bmp',
    '.ico':  'image/x-icon',
    '.cur':  'image/x-icon',

    # Professional imaging
    '.tif':  'image/tiff',
    '.tiff': 'image/tiff',
}

def convert_to_uri(
    file_path: str,
    fallback_mime: str = "image/png" #"application/octet-stream"
) -> str:
    """
    Convert a local image file to a base64-encoded data URI with the correct MIME type.

    Supports 20+ image formats (PNG, JPEG, WebP, AVIF, HEIC, SVG, BMP, ICO, TIFF, etc.).

    Args:
        file_path: Path to the image file on disk.
        fallback_mime: MIME type used when the file extension is unknown.

    Returns:
        A valid data URI string (e.g., data:image/webp;base64,...).

    Raises:
        FileNotFoundError: If the file does not exist.
        OSError: If reading the file fails.
    """
    if not os.path.isfile(file_path):
        return f"Image file not found: {file_path}"

    extension = os.path.splitext(file_path)[1].strip().lower()
    mime_type = _IMAGE_MIME_TYPES.get(extension, fallback_mime)

    if mime_type == fallback_mime and extension != ".png":
        print(f"Warning: Unknown extension '{extension}' for '{file_path}'. "
              f"Using fallback MIME type: {fallback_mime}")

    try:
        with open(file_path, "rb") as img_file:
            encoded_data = base64.b64encode(img_file.read()).decode("utf-8")
    except Exception as e:
        return f"Failed to read image file '{file_path}': {e}"

    return f"data:{mime_type};base64,{encoded_data}"






# 2 functions to assure compatibility between LLM and VLM
def is_multimodal_template(chat_template: str) -> bool:
    """
    Heuristically detect if a Jinja chat template supports multimodal input.
    """

    multimodal_markers = [
        "vision_start",
        "vision_end",
        "image_pad",
        "video_pad",
        "image_url",
        "image",
        "video",
        "<|vision_start|>",
        "<|image_pad|>",
        "<|video_pad|>",
    ]

    template_lower = chat_template.lower()

    return any(marker.lower() in template_lower for marker in multimodal_markers)

def flatten_multimodal_messages(messages, drop_images=True):
    """
    Converts structured messages into pure text format.
    Only works safely if messages contain no real multimodal content.
    """

    flat_messages = []

    for msg in messages:
        content = msg.get("content")

        # CASE 1: already string → keep
        if isinstance(content, str):
            flat_messages.append({
                "role": msg["role"],
                "content": content
            })
            continue

        # CASE 2: list of content blocks
        if isinstance(content, list):
            text_parts = []

            for item in content:
                if not isinstance(item, dict):
                    continue

                # TEXT
                if "text" in item:
                    text_parts.append(item["text"])

                # IMAGE (decide behavior)
                elif any(k in item for k in ["image", "image_url"]):
                    if not drop_images:
                        raise ValueError("Multimodal content found (image). Cannot flatten safely.")
                    # otherwise ignore images silently or mark them
                    text_parts.append("[IMAGE]")

            flat_messages.append({
                "role": msg["role"],
                "content": "\n".join(text_parts)
            })
            continue

        # CASE 3: dict-style content (your older format)
        if isinstance(content, dict):
            if "text" in content:
                flat_messages.append({
                    "role": msg["role"],
                    "content": content["text"]
                })
            elif "image_url" in content:
                if drop_images:
                    flat_messages.append({
                        "role": msg["role"],
                        "content": "[IMAGE]"
                    })
                else:
                    raise ValueError("Image found in dict content")

    return flat_messages



import os
import sys

import Orange.data
from Orange.data import StringVariable
from AnyQt.QtWidgets import QApplication
from Orange.widgets.settings import Setting
from Orange.widgets.utils.signals import Input, Output
from sentence_transformers import SentenceTransformer
from AnyQt.QtCore import QTimer


if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
    from Orange.widgets.orangecontrib.AAIT.llm import embeddings
    from Orange.widgets.orangecontrib.AAIT.utils import thread_management, base_widget, help_management
    from Orange.widgets.orangecontrib.AAIT.utils.initialize_from_ini import apply_modification_from_python_file
else:
    from orangecontrib.AAIT.llm import embeddings
    from orangecontrib.AAIT.utils import thread_management, base_widget, help_management
    from orangecontrib.AAIT.utils.initialize_from_ini import apply_modification_from_python_file


@apply_modification_from_python_file(filepath_original_widget=__file__)
class OWCreateEmbeddings(base_widget.BaseListWidget):
    name = "Create Embeddings"
    description = "Create embeddings on the column 'content' of a Table"
    category = "AAIT - LLM INTEGRATION"
    icon = "icons/owembeddings.svg"
    if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
        icon = "icons_dev/owembeddings.svg"
    gui = os.path.join(os.path.dirname(os.path.abspath(__file__)), "designer/owembeddings.ui")
    want_control_area = False
    priority = 1060

    # Settings
    selected_column_name = Setting("content")

    class Inputs:
        data = Input("Data", Orange.data.Table)
        model = Input("Model", SentenceTransformer, auto_summary=False)

    class Outputs:
        data = Output("Data", Orange.data.Table)

    @Inputs.data
    def set_data(self, in_data):
        self.data = in_data
        if self.data:
            self.var_selector.add_variables(self.data.domain)
            self.var_selector.select_variable_by_name(self.selected_column_name)
        if self.autorun:
            self.run()

    @Inputs.model
    def set_model(self, in_model):
        self.model = in_model
        if self.autorun:
            self.run()

    def __init__(self):
        super().__init__()
        # Qt Management
        self.setFixedWidth(470)
        self.setFixedHeight(450)
        #uic.loadUi(self.gui, self)

        # Data Management
        self.data = None
        self.model = None
        self.thread = None
        self.autorun = True
        self.result = None
        self.post_initialized()
        QTimer.singleShot(0, lambda: help_management.override_help_action(self))

    def run(self):
        self.error("")
        self.warning("")

        # if thread is running quit
        if self.thread is not None:
            self.thread.safe_quit()

        if self.data is None:
            self.Outputs.data.send(None)
            return

        if self.model is None:
            self.Outputs.data.send(None)
            return

        # Verification of in_data
        if not self.selected_column_name in self.data.domain:
            self.warning(f'Previously selected column "{self.selected_column_name}" does not exist in your data.')
            return

        if not isinstance(self.data.domain[self.selected_column_name], StringVariable):
            self.error('You must select a text variable.')
            return

        # Start progress bar
        self.progressBarInit()

        # Connect and start thread : main function, progress, result and finish
        # --> progress is used in the main function to track progress (with a callback)
        # --> result is used to collect the result from main function
        # --> finish is just an empty signal to indicate that the thread is finished
        self.thread = thread_management.Thread(embeddings.create_embeddings, self.data, self.model, self.selected_column_name)
        self.thread.progress.connect(self.handle_progress)
        self.thread.result.connect(self.handle_result)
        self.thread.finish.connect(self.handle_finish)
        self.thread.start()

    def handle_progress(self, value: float) -> None:
        self.progressBarSet(value)

    def handle_result(self, result):
        try:
            self.result = result
            self.Outputs.data.send(result)
        except Exception as e:
            print("An error occurred when sending out_data:", e)
            self.Outputs.data.send(None)
            return

    def handle_finish(self):
        print("Embeddings finished")
        self.progressBarFinished()

    def post_initialized(self):
        pass

if __name__ == "__main__":
    app = QApplication(sys.argv)
    my_widget = OWCreateEmbeddings()
    my_widget.show()
    if hasattr(app, "exec"):
        app.exec()
    else:
        app.exec_()



# https://orange3.readthedocs.io/projects/orange-development/en/latest/tutorial-settings.html
import os
import sys

import Orange.data
from Orange.data import ContinuousVariable, Domain
from Orange.data.table import Table
from Orange.widgets import widget
from Orange.widgets.utils.signals import Input, Output
from AnyQt.QtCore import QTimer


if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
    from Orange.widgets.orangecontrib.AAIT.optimiser import optuna_multi
    from Orange.widgets.orangecontrib.AAIT.utils.initialize_from_ini import apply_modification_from_python_file
    from Orange.widgets.orangecontrib.AAIT.utils import help_management
else:
    from orangecontrib.AAIT.optimiser import optuna_multi
    from orangecontrib.AAIT.utils.initialize_from_ini import apply_modification_from_python_file
    from orangecontrib.AAIT.utils import help_management

@apply_modification_from_python_file(filepath_original_widget=__file__)
class Optimisation(widget.OWWidget):
    name = "Optimization"
    description = "Optimization with optuna"
    icon = "icons/owoptimisation.svg"
    category = "AAIT - ALGORITHM"
    if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
        icon = "icons_dev/owoptimisation.svg"



    priority = 1020
    dossier_du_script = os.path.dirname(os.path.abspath(__file__))
    want_control_area = False
    partie_1 = ""
    partie_2 = ""
    CaptionDuLabelPourGoto = ""
    AddAcondition = 0
    liste_a_afficher_comboBox_selection_label = []

    class Inputs:
        data_search_space = Input("Search Space Limits", Orange.data.Table)
        data_number_iterations = Input("Number of iterations", Orange.data.Table)
        data_previous_study = Input("Previous Study Data", Orange.data.Table)

    class Outputs:
        data_out = Output("Data", Orange.data.Table)
        best_trial_out = Output("Best trial", Orange.data.Table)
        current_proposition_out = Output("Current proposition (preview)", Orange.data.Table)
        studies_out = Output("Last Studies ", Orange.data.Table, auto_summary=False)
        out_pointer = Output("begin of the loop", str, auto_summary=False)  # sended as the widget is created
        out_pointer1 = Output("Send study for the loop", str, auto_summary=False)  # sended as the widget is created

    @Inputs.data_number_iterations
    def set_data_number_iterations(self, dataset):
        if dataset is not None:
            self.current_data_number_iterations = dataset[0][0]
            self.compteur_iteration = 1
            # pour relancer une etude
            self.itere(score=None)

    @Inputs.data_search_space
    def set_data_search_space(self, dataset):
        if dataset is not None:
            self.current_data_search_space = dataset
            self.len_file_current_data_search_space = len(dataset)

    @Inputs.data_previous_study
    def set_data_previous_study(self, dataset):
        self.current_data_previous_study = dataset


    # on lance l'étude et définition des data table en sortie
    def launch_study(self, score):
        ligne_best_etude, data, domain, current_proposition_out = optuna_multi.launch_study(score, self.len_score, self.current_data_search_space, self.current_data_previous_study, self.stop)

        if ligne_best_etude != []:
            table_best_etude = Table.from_list(Domain([ContinuousVariable(h) for h in domain]), ligne_best_etude)
            self.Outputs.best_trial_out.send(table_best_etude)

        if current_proposition_out != []:
            self.table = Table.from_list(Domain([ContinuousVariable(h) for h in domain]), current_proposition_out)
            self.Outputs.current_proposition_out.send(self.table)

        self.current_data_previous_study = Table.from_list(Domain([ContinuousVariable(h) for h in domain]), data[3:])
        self.Outputs.studies_out.send(self.current_data_previous_study)

        if self.stop == True:
            self.stop = False
            return

    def change_last_ligne(self, data=[]):
        if data == []:
            return

        if self.current_data_previous_study is not None:
            new_ligne = []
            for i in range(len(data.domain.attributes)):
                if data[0][i] == '?':
                    new_ligne.append("?")
                else:
                    # si le pas est défini dans le fichier d'entrée on gère le pas pour éviter toute cassure avec optuna
                    if len(self.current_data_search_space) == 3:
                        new_value = round(data[0][i].value / self.current_data_search_space[2][i].value) * self.current_data_search_space[2][i].value
                        data[0][i] = new_value
                    new_ligne.append(data[0][i].value)

            replace = False
            for i in range(len(data.domain.attributes)):
                if self.current_data_previous_study[-1][i] == '?':
                    replace = True
            #modification de la dernière ligne
            if replace == True:
                self.current_data_previous_study[-1] = new_ligne

            #ajout d'une nouvelle étude
            else:
                data = []
                for i in range(len(self.current_data_previous_study)):
                    d = []
                    for j in range(len(self.current_data_previous_study.domain)):
                        d.append(self.current_data_previous_study[i][j].value)

                    data.append(d)
                data.append(new_ligne)
                self.current_data_previous_study = Table.from_list(self.current_data_previous_study.domain,
                                                                    data)

            self.Outputs.studies_out.send(self.current_data_previous_study)

    # appelé depuis un focntion tierce
    def itere(self, score=[]):
        if self.current_data_search_space is None:
            print('Erreur il n y a pas de paramètres de configuration en entree')
            return

        if self.current_data_number_iterations is None:
            if self.compteur_iteration > 10:
                self.stop = True
        else:
            if self.compteur_iteration > self.current_data_number_iterations:
                self.stop = True

        self.compteur_iteration = self.compteur_iteration + 1

        # dans le cas du chargement d'une ancienne étude
        if self.current_data_previous_study is not None:
            self.len_score = len(self.current_data_previous_study.domain) - len(self.current_data_search_space.domain)

        if score is not None:
            new_score = []
            for i in range(len(score[0])):
                if score[0][i] != '?':
                   new_score.append(score[0][i].value)
            # ajouter car le feature constructor peut garder les anciens scores
            if self.len_score is None:
                self.len_score = len(new_score)

            #if len(new_score) != self.len_score:
                #print('Erreur les nombres de scores entre les études sont différents')
                #return
            self.launch_study(new_score)

        else:
            self.launch_study(score)



    def __init__(self):

        self.current_data_standard_input = None
        self.current_data_search_space = None
        self.current_data_previous_study = None
        self.current_data_number_iterations = None
        self.etude = None
        self.stop = False
        self.len_score = None
        self.table = None
        # compteur_iteration
        self.compteur_iteration = 1
        self.len_file_current_data_search_space = 0

        # on envoie l identifiant pour pouvoir boucler
        self.Outputs.out_pointer.send(str(id(self)))

        # on envoie l identifiant pour pouvoir boucler
        self.Outputs.out_pointer1.send(str(id(self)))

        super().__init__()  # je pense que c'est pour garder les config entre deux executions


        self.setFixedWidth(480)
        self.setFixedHeight(354)
        self.setAutoFillBackground(True)
        QTimer.singleShot(0, lambda: help_management.override_help_action(self))


if __name__ == "__main__":
    from AnyQt.QtWidgets import QApplication
    app = QApplication(sys.argv)
    mon_objet = Optimisation()
    mon_objet.show()
    if hasattr(app, "exec"):
        app.exec()
    else:
        app.exec_()






from orangecanvas.localization import Translator  # pylint: disable=wrong-import-order
_tr = Translator("Orange", "biolab.si", "Orange")
del Translator
from typing import Optional, Callable, List, Union, Dict, Tuple
from collections import namedtuple
from functools import singledispatch

import numpy as np

from AnyQt.QtCore import QTimer, Qt, QSortFilterProxyModel, QSize, QDateTime, \
    QModelIndex, Signal, QPoint, QRect, QEvent
from AnyQt.QtGui import QStandardItemModel, QStandardItem, QIcon, QPainter, \
    QColor, QValidator
from AnyQt.QtWidgets import QLineEdit, QTableView, QSlider, \
    QComboBox, QStyledItemDelegate, QWidget, QDateTimeEdit, QHBoxLayout, \
    QDoubleSpinBox, QSizePolicy, QStyleOptionViewItem, QLabel, QMenu, QAction

from orangewidget.gui import Slider

from Orange.data import DiscreteVariable, ContinuousVariable, \
    TimeVariable, Table, StringVariable, Variable, Domain
from Orange.data.util import get_unique_names
from Orange.widgets import gui
from Orange.widgets.utils.itemmodels import TableModel
from Orange.widgets.utils.widgetpreview import WidgetPreview
from Orange.widgets.widget import OWWidget, Input, Output, Msg
from Orange.widgets.settings import Setting
import os
if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
    from Orange.widgets.orangecontrib.AAIT.utils import widget_positioning, help_management
else:
    from orangecontrib.AAIT.utils import widget_positioning, help_management

VariableRole = next(gui.OrangeUserRole)
ValuesRole = next(gui.OrangeUserRole)
ValueRole = next(gui.OrangeUserRole)


class VariableEditor(QWidget):
    valueChanged = Signal(float)

    def __init__(self, parent: QWidget, callback: Callable):
        super().__init__(parent)
        layout = QHBoxLayout()
        layout.setContentsMargins(6, 0, 6, 0)
        layout.setAlignment(Qt.AlignLeft)
        self.setLayout(layout)
        self.valueChanged.connect(callback)

    @property
    def value(self) -> Union[int, float, str]:
        return NotImplemented

    @value.setter
    def value(self, value: Union[float, str]):
        raise NotImplementedError

    def sizeHint(self):
        return QSize(super().sizeHint().width(), 40)


class DiscreteVariableEditor(VariableEditor):
    def __init__(self, parent: QWidget, items: Tuple[str], callback: Callable):
        super().__init__(parent, callback)
        self._combo = QComboBox(
            parent,
            maximumWidth=180,
            sizePolicy=QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        )
        self._combo.addItems(items + ("?",))
        self._combo.currentIndexChanged.connect(self.__on_index_changed)
        self.layout().addWidget(self._combo)

    @property
    def value(self) -> Union[int, float]:
        return self._map_to_var_values()

    @value.setter
    def value(self, value: float):
        if np.isnan(value):
            value = self._combo.model().rowCount() - 1
        assert value == int(value)
        self._combo.setCurrentIndex(int(value))

    def __on_index_changed(self):
        self.valueChanged.emit(self._map_to_var_values())

    def _map_to_var_values(self) -> Union[int, float]:
        n_values = self._combo.model().rowCount() - 1
        current = self._combo.currentIndex()
        return current if current < n_values else np.nan


class ContinuousVariableEditor(VariableEditor):
    MAX_FLOAT = 2147483647

    def __init__(self, parent: QWidget, variable: ContinuousVariable,
                 min_value: float, max_value: float, callback: Callable):
        super().__init__(parent, callback)

        if np.isnan(min_value) or np.isnan(max_value):
            raise ValueError("Min/Max cannot be NaN.")

        n_decimals = variable.number_of_decimals
        abs_max = max(abs(min_value), max_value)
        if abs_max * 10 ** n_decimals > self.MAX_FLOAT:
            n_decimals = int(np.log10(self.MAX_FLOAT / abs_max))

        self._value: float = min_value
        self._n_decimals: int = n_decimals
        self._min_value: float = self.__round_value(min_value)
        self._max_value: float = self.__round_value(max_value)

        sp_spin = QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        sp_spin.setHorizontalStretch(1)
        sp_slider = QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        sp_slider.setHorizontalStretch(5)
        sp_edit = QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        sp_edit.setHorizontalStretch(1)

        class DoubleSpinBox(QDoubleSpinBox):
            def sizeHint(self) -> QSize:
                size: QSize = super().sizeHint()
                return QSize(size.width(), size.height() + 2)

            def validate(self, text: str, pos: int) -> Tuple[int, str, int]:
                state, text, pos = super().validate(text, pos)
                if text == "":
                    state = QValidator.Acceptable
                return state, text, pos

            def textFromValue(self, value):
                if not np.isfinite(value):
                    return "?"
                return super().textFromValue(value)

        self._spin = DoubleSpinBox(
            parent,
            value=self._min_value,
            minimum=-np.inf,
            maximum=np.inf,
            singleStep=10 ** (-self._n_decimals),
            decimals=self._n_decimals,
            minimumWidth=70,
            sizePolicy=sp_spin,
        )
        self._slider = Slider(
            parent,
            minimum=self.__map_to_slider(self._min_value),
            maximum=self.__map_to_slider(self._max_value),
            singleStep=1,
            orientation=Qt.Horizontal,
            sizePolicy=sp_slider,
        )
        self._label_min = QLabel(
            parent,
            text=variable.repr_val(min_value),
            alignment=Qt.AlignRight,
            minimumWidth=60,
            sizePolicy=sp_edit,
        )
        self._label_max = QLabel(
            parent,
            text=variable.repr_val(max_value),
            alignment=Qt.AlignLeft,
            minimumWidth=60,
            sizePolicy=sp_edit,
        )

        self._slider.valueChanged.connect(self._apply_slider_value)
        self._spin.valueChanged.connect(self._apply_spin_value)

        self.layout().addWidget(self._spin)
        self.layout().addWidget(self._label_min)
        self.layout().addWidget(self._slider)
        self.layout().addWidget(self._label_max)

        self.setFocusProxy(self._spin)

        def deselect():
            self._spin.lineEdit().deselect()
            try:
                self._spin.lineEdit().selectionChanged.disconnect(deselect)
            except TypeError:
                pass

        self._spin.lineEdit().selectionChanged.connect(deselect)
        self._slider.installEventFilter(self)
        self._spin.installEventFilter(self)

    @property
    def value(self) -> float:
        return self.__round_value(self._value)

    @value.setter
    def value(self, value: float):
        if self._value is None or self.__round_value(value) != self.value:
            self._value = value
            self.valueChanged.emit(self.value)
            self._spin.setValue(self.value)
            slider_value = self.__map_to_slider(self.value)
            self._value = self.__map_from_slider(slider_value)
            self._slider.setValue(slider_value)
            self._value = value

    def _apply_slider_value(self):
        self.value = self.__map_from_slider(self._slider.value())

    def _apply_spin_value(self):
        value = self._spin.value()
        self.value = value if np.isfinite(value) else np.nan

    def __round_value(self, value):
        return round(value, self._n_decimals)

    def __map_to_slider(self, value: float) -> int:
        value = min(self._max_value, max(self._min_value, value))
        return round(value * 10 ** self._n_decimals)

    def __map_from_slider(self, value: int) -> float:
        return value * 10 ** (-self._n_decimals)

    def eventFilter(self, obj: Union[QSlider, QDoubleSpinBox], event: QEvent) -> bool:
        if event.type() == QEvent.Wheel:
            return True
        return super().eventFilter(obj, event)


class StringVariableEditor(VariableEditor):
    valueChanged = Signal()

    def __init__(self, parent: QWidget, callback: Callable):
        super().__init__(parent, callback)
        self._edit = QLineEdit(
            parent,
            sizePolicy=QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        )
        self._edit.textChanged.connect(self.valueChanged)
        self.layout().addWidget(self._edit)
        self.setFocusProxy(self._edit)

    @property
    def value(self) -> str:
        return self._edit.text()

    @value.setter
    def value(self, value: str):
        self._edit.setText(value)


class TimeVariableEditor(VariableEditor):
    DATE_FORMAT = "yyyy-MM-dd"
    TIME_FORMAT = "hh:mm:ss"

    def __init__(self, parent: QWidget, variable: TimeVariable, callback: Callable):
        super().__init__(parent, callback)
        self._value: float = 0
        self._variable: TimeVariable = variable

        if variable.have_date and not variable.have_time:
            self._format = TimeVariableEditor.DATE_FORMAT
        elif not variable.have_date and variable.have_time:
            self._format = TimeVariableEditor.TIME_FORMAT
        else:
            self._format = (f"{TimeVariableEditor.DATE_FORMAT} "
                            f"{TimeVariableEditor.TIME_FORMAT}")

        class DateTimeEdit(QDateTimeEdit):
            def sizeHint(self) -> QSize:
                size: QSize = super().sizeHint()
                return QSize(size.width(), size.height() + 2)

        self._edit = DateTimeEdit(
            parent,
            dateTime=self.__map_to_datetime(self._value),
            displayFormat=self._format,
            sizePolicy=QSizePolicy(QSizePolicy.Maximum, QSizePolicy.Fixed)
        )
        self._edit.dateTimeChanged.connect(self._apply_edit_value)
        self.layout().addWidget(self._edit)
        self.setFocusProxy(self._edit)
        self._edit.installEventFilter(self)

    @property
    def value(self) -> float:
        return self._value

    @value.setter
    def value(self, value: float):
        if value != self.value:
            self._value = value
            self.valueChanged.emit(self.value)
            self._edit.setDateTime(self.__map_to_datetime(self.value))

    def _apply_edit_value(self):
        self.value = self.__map_from_datetime(self._edit.dateTime())

    def __map_from_datetime(self, date_time: QDateTime) -> float:
        return self._variable.to_val(date_time.toString(self._format))

    def __map_to_datetime(self, value: float) -> QDateTime:
        return QDateTime.fromString(self._variable.repr_val(value), self._format)

    def eventFilter(self, obj: QDateTimeEdit, event: QEvent) -> bool:
        if event.type() == QEvent.Wheel:
            return True
        return super().eventFilter(obj, event)


class VariableDelegate(QStyledItemDelegate):
    def paint(self, painter: QPainter, option: QStyleOptionViewItem, index: QModelIndex):
        self.parent().view.openPersistentEditor(index)
        super().paint(painter, option, index)

    def createEditor(self, parent: QWidget, _: QStyleOptionViewItem,
                     index: QModelIndex) -> VariableEditor:
        variable = index.data(VariableRole)
        values = index.data(ValuesRole)
        return _create_editor(variable, values, parent, self._commit_data)

    def _commit_data(self):
        editor = self.sender()
        assert isinstance(editor, VariableEditor)
        self.commitData.emit(editor)

    # pylint: disable=no-self-use
    def setEditorData(self, editor: VariableEditor, index: QModelIndex):
        editor.value = index.model().data(index, ValueRole)

    # pylint: disable=no-self-use
    def setModelData(self, editor: VariableEditor,
                     model: QSortFilterProxyModel, index: QModelIndex):
        model.setData(index, editor.value, ValueRole)

    # pylint: disable=no-self-use
    def updateEditorGeometry(self, editor: VariableEditor,
                             option: QStyleOptionViewItem, _: QModelIndex):
        rect: QRect = option.rect
        if isinstance(editor, ContinuousVariableEditor):
            width = editor.sizeHint().width()
            if width > rect.width():
                rect.setWidth(width)
        editor.setGeometry(rect)

    # pylint: disable=no-self-use
    def sizeHint(self, _: QStyleOptionViewItem, index: QModelIndex) -> QSize:
        return _create_editor(index.data(role=VariableRole), np.array([0]),
                              None, lambda: 1).sizeHint()


@singledispatch
def _create_editor(*_) -> VariableEditor:
    raise NotImplementedError


@_create_editor.register(DiscreteVariable)
def _(variable: DiscreteVariable, _: np.ndarray,
      parent: QWidget, callback: Callable) -> DiscreteVariableEditor:
    return DiscreteVariableEditor(parent, variable.values, callback)


@_create_editor.register(ContinuousVariable)
def _(variable: ContinuousVariable, values: np.ndarray,
      parent: QWidget, callback: Callable) -> ContinuousVariableEditor:
    return ContinuousVariableEditor(parent, variable, np.nanmin(values),
                                    np.nanmax(values), callback)


@_create_editor.register(StringVariable)
def _(_: StringVariable, __: np.ndarray, parent: QWidget,
      callback: Callable) -> StringVariableEditor:
    return StringVariableEditor(parent, callback)


@_create_editor.register(TimeVariable)
def _(variable: TimeVariable, _: np.ndarray,
      parent: QWidget, callback: Callable) -> TimeVariableEditor:
    return TimeVariableEditor(parent, variable, callback)


def majority(values: np.ndarray) -> Union[int, float]:
    if all(np.isnan(values)):
        return np.nan
    return np.bincount(values[~np.isnan(values)].astype(int)).argmax()


def disc_random(values: np.ndarray) -> int:
    return np.random.randint(low=np.nanmin(values), high=np.nanmax(values) + 1)


def cont_random(values: np.ndarray) -> float:
    return np.random.uniform(low=np.nanmin(values), high=np.nanmax(values))


class VariableItemModel(QStandardItemModel):
    dataHasNanColumn = Signal()

    # pylint: disable=dangerous-default-value
    def set_data(self, data: Table, saved_values={}):
        domain = data.domain
        variables = [(TableModel.Attribute, a) for a in domain.attributes] + \
                    [(TableModel.ClassVar, c) for c in domain.class_vars] + \
                    [(TableModel.Meta, m) for m in domain.metas]
        for place, variable in variables:
            if variable.is_primitive():
                values = data.get_column(variable)
                if all(np.isnan(values)):
                    self.dataHasNanColumn.emit()
                    continue
            else:
                values = np.array([])
            color = TableModel.ColorForRole.get(place)
            self._add_row(variable, values, color, saved_values.get(variable.name))

    def _add_row(self, variable: Variable, values: np.ndarray, color: QColor,
                 saved_value: Optional[Union[int, float, str]]):
        var_item = QStandardItem()
        var_item.setData(variable.name, Qt.DisplayRole)
        var_item.setToolTip(variable.name)
        var_item.setIcon(self._variable_icon(variable))
        var_item.setEditable(False)
        if color:
            var_item.setBackground(color)

        control_item = QStandardItem()
        control_item.setData(variable, VariableRole)
        control_item.setData(values, ValuesRole)
        if color:
            control_item.setBackground(color)

        value = self._default_for_variable(variable, values)
        if saved_value is not None and not \
                (variable.is_discrete and saved_value >= len(variable.values)):
            value = saved_value
        control_item.setData(value, ValueRole)

        self.appendRow([var_item, control_item])

    @staticmethod
    def _default_for_variable(variable: Variable, values: np.ndarray) \
            -> Union[float, int, str]:
        if variable.is_continuous:
            return round(np.nanmedian(values), variable.number_of_decimals)
        elif variable.is_discrete:
            return majority(values)
        elif variable.is_string:
            return ""
        else:
            raise NotImplementedError

    @staticmethod
    def _variable_icon(variable: Variable) -> QIcon:
        if variable.is_discrete:
            return gui.attributeIconDict[1]
        elif variable.is_time:
            return gui.attributeIconDict[4]
        elif variable.is_continuous:
            return gui.attributeIconDict[2]
        elif variable.is_string:
            return gui.attributeIconDict[3]
        else:
            return gui.attributeIconDict[-1]


class OWCreateInstance(OWWidget):
    name = "Auto show Create Instance"
    description = "Interactively create a data instance from sample dataset."
    icon = "icons/CreateInstance.svg"
    if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
        icon = "icons_dev/CreateInstance.svg"
    category = "AAIT - TOOLBOX"
    keywords = "create instance, simulator"
    priority = 2310

    class Inputs:
        data = Input("Data", Table)
        reference = Input("Reference", Table)
        input_autoshow = Input("AutoShowConfiguration", str, auto_summary=False)

    class Outputs:
        data = Output("Data", Table)

    class Information(OWWidget.Information):
        nans_removed = Msg("Variables with only missing values were removed from the list.")

    str_WidgetPositionning: str = Setting("None")

    want_main_area = False
    BUTTONS = ["Median", "Mean", "Random", "Input"]
    ACTIONS = ["median", "mean", "random", "input"]
    HEADER = [["name", "Variable"],
              ["variable", "Value"]]
    Header = namedtuple(
        "header", [tag for tag, _ in HEADER]
    )(*range(len(HEADER)))

    values: Dict[str, Union[float, str]] = Setting({}, schema_only=True)
    append_to_data = Setting(True)
    auto_commit = Setting(True)
    auto_apply_reference = Setting(True)

    def __init__(self):
        super().__init__()
        self.data: Optional[Table] = None
        self.reference: Optional[Table] = None

        # When True, dataChanged from the model does NOT trigger _do_commit.
        # Set before every programmatic model update, cleared in finally.
        self._silent_update = False

        self.filter_edit = QLineEdit(textChanged=self.__filter_edit_changed,
                                     placeholderText="Filter...")
        self.view = QTableView(sortingEnabled=True,
                               contextMenuPolicy=Qt.CustomContextMenu,
                               selectionMode=QTableView.NoSelection)
        self.view.customContextMenuRequested.connect(self.__menu_requested)
        self.view.setItemDelegateForColumn(
            self.Header.variable, VariableDelegate(self)
        )
        self.view.verticalHeader().hide()
        self.view.horizontalHeader().setStretchLastSection(True)
        self.view.horizontalHeader().setMaximumSectionSize(350)

        self.model = VariableItemModel(self)
        self.model.setHorizontalHeaderLabels([x for _, x in self.HEADER])
        self.model.dataChanged.connect(self.__table_data_changed)
        self.model.dataHasNanColumn.connect(self.Information.nans_removed)
        self.proxy_model = QSortFilterProxyModel()
        self.proxy_model.setFilterKeyColumn(-1)
        self.proxy_model.setFilterCaseSensitivity(Qt.CaseInsensitive)
        self.proxy_model.setSourceModel(self.model)
        self.view.setModel(self.proxy_model)

        vbox = gui.vBox(self.controlArea, box=True)
        vbox.layout().addWidget(self.filter_edit)
        vbox.layout().addWidget(self.view)

        box = gui.hBox(vbox, objectName="buttonBox")
        gui.rubber(box)
        for name, action in zip(self.BUTTONS, self.ACTIONS):
            gui.button(
                box, self, name,
                lambda *args, fun=action: self._initialize_values(fun),
                autoDefault=False
            )
        gui.rubber(box)

        gui.checkBox(self.buttonsArea, self, "append_to_data",
                     "Append this instance to input data",
                     callback=self._do_commit_if_auto)

        gui.checkBox(
            self.buttonsArea, self, "auto_apply_reference",
            self.tr("Auto-apply Reference to UI"),
            callback=self._on_auto_apply_reference_changed
        )

        gui.rubber(self.buttonsArea)

        # Apply button (always enabled) + "Apply automatically" checkbox
        apply_box = gui.hBox(self.buttonsArea)
        self._apply_button = gui.button(
            apply_box, self, "Apply",
            callback=self._do_commit,   # appel direct, pas via gui.deferred
            autoDefault=False
        )
        gui.checkBox(
            apply_box, self, "auto_commit", "Apply automatically",
            callback=self._on_auto_commit_changed
        )

        self.settingsAboutToBePacked.connect(self.pack_settings)
        widget_positioning.show_and_adjust_at_opening(self, str(self.str_WidgetPositionning))
        QTimer.singleShot(0, lambda: help_management.override_help_action(self))

    # ------------------------------------------------------------------
    # Core output method — plain Python, no Orange deferred machinery
    # ------------------------------------------------------------------

    def _do_commit(self):
        """Build and send the output. Always executes immediately."""
        output_data = None
        if self.data:
            output_data = self._create_data_from_values()
            if self.append_to_data:
                output_data = self._append_to_data(output_data)
        self.Outputs.data.send(output_data)

    def _do_commit_if_auto(self):
        """Send only when 'Apply automatically' is checked."""
        if self.auto_commit:
            self._do_commit()

    # ------------------------------------------------------------------
    # UI callbacks
    # ------------------------------------------------------------------

    def _on_auto_commit_changed(self):
        if self.auto_commit:
            self._do_commit()

    def _on_auto_apply_reference_changed(self):
        if self.auto_apply_reference:
            self._apply_reference_to_ui()

    def __filter_edit_changed(self):
        self.proxy_model.setFilterFixedString(self.filter_edit.text().strip())
        if self.auto_apply_reference and self.reference is not None:
            self._apply_reference_to_ui()

    def __table_data_changed(self):
        """Fired when the user manually moves a slider / combo.
        Suppressed during all programmatic updates via _silent_update."""
        if not self._silent_update and self.auto_commit:
            self._do_commit()

    def __menu_requested(self, point: QPoint):
        index = self.view.indexAt(point)
        model: QSortFilterProxyModel = index.model()
        source_index = model.mapToSource(index)
        menu = QMenu(self)
        for action in self._create_actions(source_index):
            menu.addAction(action)
        menu.popup(self.view.viewport().mapToGlobal(point))

    def _create_actions(self, index: QModelIndex) -> List[QAction]:
        actions = []
        for name in self.ACTIONS:
            action = QAction(name.capitalize(), self)
            action.triggered.connect(
                lambda *args, fun=name: self._initialize_values(fun, [index])
            )
            actions.append(action)
        return actions

    # ------------------------------------------------------------------
    # Reference application — always silent, never commits
    # ------------------------------------------------------------------

    def _apply_reference_to_ui(self, indices: List[QModelIndex] = None):
        """Update UI values from reference row. Never triggers a commit."""
        if not self.data or self.reference is None or len(self.reference) < 1:
            return

        ref_table = self.reference
        self._silent_update = True
        try:
            if indices is not None:
                source_indexes = indices
            else:
                source_indexes = [
                    self.proxy_model.mapToSource(
                        self.proxy_model.index(prow, self.Header.variable)
                    )
                    for prow in range(self.proxy_model.rowCount())
                ]

            for sindex in source_indexes:
                var = self.model.data(sindex, VariableRole)
                if var is None or var.name not in ref_table.domain:
                    continue
                ref_var = ref_table.domain[var.name]
                if (var.is_discrete != ref_var.is_discrete) or \
                        (var.is_continuous != ref_var.is_continuous) or \
                        (var.is_string != ref_var.is_string) or \
                        (var.is_time != ref_var.is_time):
                    continue
                col = ref_table.get_column(ref_var)
                value = col[0] if len(col) else np.nan
                self.model.setData(sindex, value, ValueRole)
        finally:
            self._silent_update = False

    # ------------------------------------------------------------------
    # Button actions — silent UI update, no automatic commit
    # ------------------------------------------------------------------

    def _initialize_values(self, fun: str, indices: List[QModelIndex] = None):
        cont_fun = {"median": np.nanmedian,
                    "mean": np.nanmean,
                    "random": cont_random,
                    "input": np.nanmean}.get(fun, NotImplemented)
        disc_fun = {"median": majority,
                    "mean": majority,
                    "random": disc_random,
                    "input": majority}.get(fun, NotImplemented)

        if not self.data or (fun == "input" and not self.reference):
            return

        if fun == "input":
            self._apply_reference_to_ui(indices)
            return

        self._silent_update = True
        try:
            if indices is None:
                source_indexes = [
                    self.proxy_model.mapToSource(
                        self.proxy_model.index(r, self.Header.variable)
                    )
                    for r in range(self.proxy_model.rowCount())
                ]
            else:
                source_indexes = indices

            for sindex in source_indexes:
                variable = self.model.data(sindex, VariableRole)
                values = self.model.data(sindex, ValuesRole)

                if variable.is_continuous:
                    value = cont_fun(values)
                    value = round(value, variable.number_of_decimals)
                elif variable.is_discrete:
                    value = disc_fun(values)
                elif variable.is_string:
                    value = ""
                else:
                    raise NotImplementedError

                self.model.setData(sindex, value, ValueRole)
        finally:
            self._silent_update = False

    # ------------------------------------------------------------------
    # Inputs — always silent, no automatic commit on arrival
    # ------------------------------------------------------------------

    @Inputs.data
    def set_data(self, data: Table):
        self.data = data
        self._silent_update = True
        try:
            self._set_model_data()
            if self.auto_apply_reference and self.reference is not None:
                self._apply_reference_to_ui()
        finally:
            self._silent_update = False

    def _set_model_data(self):
        self.Information.nans_removed.clear()
        self.model.removeRows(0, self.model.rowCount())
        if not self.data:
            return
        self.model.set_data(self.data, self.values)
        self.values = {}
        self.view.horizontalHeader().setStretchLastSection(False)
        self.view.resizeColumnsToContents()
        self.view.resizeRowsToContents()
        self.view.horizontalHeader().setStretchLastSection(True)

    @Inputs.reference
    def set_reference(self, data: Table):
        self.reference = data
        if self.auto_apply_reference:
            self._apply_reference_to_ui()

    @Inputs.input_autoshow
    def set_input_autoshow(self, le_str):
        if le_str is not None:
            self.str_WidgetPositionning = str(le_str)

    # ------------------------------------------------------------------
    # Output builders
    # ------------------------------------------------------------------

    def _create_data_from_values(self) -> Table:
        data = Table.from_domain(self.data.domain, 1)
        with data.unlocked():
            data.name = "created"
            if data.X.size:
                data.X[:] = np.nan
            if data.Y.size:
                data.Y[:] = np.nan
            for i, m in enumerate(self.data.domain.metas):
                data.metas[:, i] = "" if m.is_string else np.nan
            values = self._get_values()
            for var_name, value in values.items():
                data[:, var_name] = value
        return data

    def _append_to_data(self, instance: Table) -> Table:
        assert self.data
        assert len(instance) == 1
        source_label = "__source_widget"

        data = Table.concatenate([self.data, instance], axis=0)
        domain = self.data.domain
        with data.unlocked():
            for attrs, part in ((domain.attributes, data.X),
                                (domain.class_vars, data.Y.reshape(len(data), -1)),
                                (domain.metas, data.metas)):
                for idx, var in enumerate(attrs):
                    if var.attributes.get(source_label) == OWCreateInstance:
                        part[-1, idx] = 1
                        return data

        name = get_unique_names(self.data.domain, "Source ID")
        var = DiscreteVariable(name, values=(self.data.name, instance.name))
        var.attributes[source_label] = OWCreateInstance
        domain = Domain(domain.attributes, domain.class_vars,
                        domain.metas + (var,))
        data = data.transform(domain)
        with data.unlocked(data.metas):
            data.metas[: len(self.data), -1] = 0
            data.metas[len(self.data):, -1] = 1
        return data

    def _get_values(self) -> Dict[str, Union[str, float]]:
        values = {}
        for row in range(self.model.rowCount()):
            index = self.model.index(row, self.Header.variable)
            values[self.model.data(index, VariableRole).name] = \
                self.model.data(index, ValueRole)
        return values

    # ------------------------------------------------------------------
    # Misc
    # ------------------------------------------------------------------

    def send_report(self):
        if not self.data:
            return
        self.report_domain("Input", self.data.domain)
        self.report_domain("Output", self.data.domain)
        items = []
        values: Dict = self._get_values()
        for var in self.data.domain.variables + self.data.domain.metas:
            val = values.get(var.name, np.nan)
            if var.is_primitive():
                val = var.repr_val(val)
            items.append([f"{var.name}:", val])
        self.report_table("Values", items)

    @staticmethod
    def sizeHint():
        return QSize(600, 500)

    def pack_settings(self):
        self.values: Dict[str, Union[str, float]] = self._get_values()


if __name__ == "__main__":  # pragma: no cover
    table = Table("housing")
    WidgetPreview(OWCreateInstance).run(set_data=table,
                                        set_reference=table[:1])



import os
import sys

from AnyQt.QtWidgets import QApplication, QComboBox
from AnyQt.QtCore import QTimer, Qt
from AnyQt.QtGui import QIcon
from Orange.widgets import widget
from Orange.widgets.utils.signals import Output, Input
from Orange.widgets.settings import Setting
from AnyQt.QtWidgets import QCheckBox
import Orange

if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
    from Orange.widgets.orangecontrib.AAIT.utils import help_management
    from Orange.widgets.orangecontrib.AAIT.utils import SimpleDialogQt
    from Orange.widgets.orangecontrib.AAIT.llm.okNokGpu import has_gpu_with_min_vram,auto_choose_model,get_model_names
    from Orange.widgets.orangecontrib.AAIT.utils.import_uic import uic
    from Orange.widgets.orangecontrib.AAIT.utils.MetManagement import GetFromRemote, get_local_store_path,ensure_file_exists_recursive
    from Orange.widgets.orangecontrib.AAIT.utils.initialize_from_ini import apply_modification_from_python_file
else:
    from orangecontrib.AAIT.utils import help_management
    from orangecontrib.AAIT.utils import SimpleDialogQt
    from orangecontrib.AAIT.llm.okNokGpu import has_gpu_with_min_vram,auto_choose_model,get_model_names
    from orangecontrib.AAIT.utils.import_uic import uic
    from orangecontrib.AAIT.utils.MetManagement import GetFromRemote, get_local_store_path,ensure_file_exists_recursive
    from orangecontrib.AAIT.utils.initialize_from_ini import apply_modification_from_python_file




@apply_modification_from_python_file(filepath_original_widget=__file__)
class OWLargeLanguageModel(widget.OWWidget):
    name = "Large Language Model"
    description = "Load a large language model (Qwen, Gemma, Phi...) for Engine LLM."
    category = "AAIT - MODELS"
    icon = "icons/owlargelanguagemodel.svg"
    if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
        icon = "icons_dev/owlargelanguagemodel.svg"
    gui = os.path.join(os.path.dirname(os.path.abspath(__file__)), "designer/owlargelanguagemodel.ui")
    priority=1093
    want_control_area = False
    autochoose_llm: str = Setting("False")
    # Settings
    model_name = Setting("Qwen3.5 9B Q6")

    class Inputs:
        data = Input("path", Orange.data.Table, auto_summary=False)

    class Outputs:
        out_model_path = Output("Model", str, auto_summary=False)


    @Inputs.data
    def set_data(self, in_data):
        self.data = in_data
        self.run()


    def __init__(self):
        super().__init__()
        # Qt Management
        self.setFixedWidth(470)
        self.setFixedHeight(300)
        uic.loadUi(self.gui, self)
        ## Combobox for model choice
        self.combobox_model = self.findChild(QComboBox, "comboBox")
        ## checboc for model autochoose_llm
        self.checkBox=self.findChild(QCheckBox, "checkBox")
        if str(self.autochoose_llm)=="False":
            self.checkBox.setChecked(False)
        else:
            self.checkBox.setChecked(True)


        self.data = None

        self.combobox_model.addItems(get_model_names().keys())
        self.display_local_models()
        self.combobox_model.setEnabled(True)
        if self.autochoose_llm=="True":
            self.model_name=auto_choose_model()
            self.combobox_model.setEnabled(False)
        self.combobox_model.setCurrentIndex(self.combobox_model.findText(self.model_name))
        self.combobox_model.currentTextChanged.connect(self.on_model_changed)
        self.checkBox.clicked.connect(self.update_checkbox)
        # Run
        self.run()
        QTimer.singleShot(0, lambda: help_management.override_help_action(self))



    def update_checkbox(self):
        if self.checkBox.isChecked():
            self.autochoose_llm="True"
            self.combobox_model.setEnabled(False)
        else:
            self.autochoose_llm = "False"
            self.combobox_model.setEnabled(True)
        self.run()

    def on_model_changed(self, text):
        self.model_name = text
        self.run()

    def display_local_models(self):
        local_store_path = get_local_store_path()

        # Prepare your icons
        icon_check = QIcon(os.path.join(os.path.dirname(os.path.abspath(__file__)), "icons/green_check.svg"))  # green check
        icon_down_body = QIcon(os.path.join(os.path.dirname(os.path.abspath(__file__)), "icons/blue_down_arrow.svg"))  # blue arrow
        model_names=get_model_names()
        for i in range(self.combobox_model.count()):
            item = self.combobox_model.model().item(i)
            filename = model_names[item.text()]
            path = os.path.join(local_store_path, "Models", "NLP", filename)
            model_path = str(path)
            list_tr = [model_path]
            if ensure_file_exists_recursive(list_tr):
                item.setForeground(Qt.black)
                item.setIcon(icon_check)
            else:
                item.setForeground(Qt.gray)
                item.setIcon(icon_down_body)


    def run(self):
        self.error("")
        self.warning("")

        if self.autochoose_llm=="True":
            self.model_name=auto_choose_model()
            self.warning("autochoose llm activated : "+str(self.model_name))

        if self.data is not None:
            if not "path" in self.data.domain:
                self.error("Requires a 'path' variable")
                return
            else:
                self.warning("")
                self.model_name = self.data[0]["path"].value.replace('"', "").replace("'", "")
                self.Outputs.out_model_path.send(self.model_name)
                return

        # Get the model path based on the selected name
        model_names=get_model_names()
        try:
            filename = model_names[self.model_name]
        except:
            filename=""
        model_path = os.path.join(get_local_store_path(), "Models", "NLP", filename)
        model_path=str(model_path)
        list_tr=[model_path]
        # Verify if model exists
        if not ensure_file_exists_recursive(list_tr):
            if not SimpleDialogQt.BoxYesNo("Model isn't in your computer. Do you want to download it from AAIT store ?"):
                self.error("Model is not on your computer.")
                return
            try:
                if 0 != GetFromRemote(self.model_name):
                    self.error("Model is not on your computer.")
                    return
                ensure_file_exists_recursive(list_tr)
            except Exception as e:
                SimpleDialogQt.BoxError(f"Unable to get the Model ({e})")
                self.error("Model is not on your computer.")
                return
        model_path=list_tr[0]
        self.Outputs.out_model_path.send(model_path)


if __name__ == "__main__":
    app = QApplication(sys.argv)
    my_widget = OWLargeLanguageModel()
    my_widget.show()
    if hasattr(app, "exec"):
        app.exec()
    else:
        app.exec_()
